import bpy
import numpy as np


class AnnotationInterface:
    """
    Base class for Annotation Interface

    AnnotationManager creates a dynamic AnnotationInterface for every instance
    of an annotation. All the annotation specific inputs and the common
    annotation params are exposed though this interface for the API, GUI and the
    annotation drawing code.

    """

    # py annotations of common params for code complete
    # these are bound dynamically along with any annotation inputs
    visible: bool
    text_font: str
    text_color: np.ndarray | list | tuple
    text_size: int
    text_align: str
    text_rotation: float
    text_vspacing: float
    text_depth: bool
    text_falloff: float
    text_offset_x: float
    text_offset_y: float
    line_mode: str
    line_color: np.ndarray | list | tuple
    line_width: float
    line_arrow_size: float
    line_pointer_length: float
    mesh_wireframe: bool
    mesh_thickness: float
    mesh_color: np.ndarray | list | tuple
    mesh_material: str | bpy.types.Material
    mesh_shade_smooth: bool

    def __init__(self, instance):
        self._instance = instance



from abc import ABCMeta
from enum import Enum
from typing import List
import bpy
from databpy import (
    BlenderObject,
)
from ..blender import utils as blender_utils
from ..nodes import nodes
from ..nodes.geometry import (
    GeometryNodeInterFace,
    style_interfaces_from_tree,
)
from .utilities import BoolObjectMNProperty, _get_gn_modifier


# create a EntityType enum
# These should match the entity_type EnumProperty in props.py
class EntityType(Enum):
    MD = "md"
    MD_OXDNA = "md-oxdna"
    MD_STREAMING = "md-streaming"
    MOLECULE = "molecule"
    ENSEMBLE = "ensemble"
    DENSITY = "density"
    ENSEMBLE_STAR = "ensemble-star"
    ENSEMBLE_CELLPACK = "ensemble-cellpack"


bpy.types.NodesModifier


class MolecularEntity(
    BlenderObject,
    metaclass=ABCMeta,
):
    update_with_scene = BoolObjectMNProperty("update_with_scene")

    def __init__(self) -> None:
        super().__init__(obj=None)
        self._register_with_session()
        self._world_scale = 0.01

    @property
    def node_group(self) -> bpy.types.NodeTree | None:
        if "MolecularNodes" in self.object.modifiers:
            return _get_gn_modifier(self.object, "MolecularNodes").node_group
        return None

    @property
    def tree(self) -> bpy.types.GeometryNodeTree:
        mod: bpy.types.NodesModifier = self.object.modifiers["MolecularNodes"]  # type: ignore
        if mod is None:
            raise ValueError(
                f"Unable to get MolecularNodes modifier for {self.object}, modifiers: {list(self.object.modifiers)}"
            )
        return mod.node_group  # type: ignore

    @property
    def styles(self) -> List[GeometryNodeInterFace]:
        """
        Get the styles in the tree.
        """
        return style_interfaces_from_tree(self.tree)

    def _register_with_session(self) -> None:
        bpy.context.scene.MNSession.register_entity(self)  # type: ignore

    def set_frame(self, frame: int) -> None:
        """
        Update the underlying data to correspond to changes in the scene frame.

        This method should be implemented by subclasses to update the entity's
        data based on the given frame number. This can include updating positions
        and performing other necessary calculations.

        Args:
            frame (int): The frame number to update the entity's data to.

        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def _setup_modifiers(self):
        """
        Create the modifiers for the molecule.
        """
        self.object.modifiers.new("MolecularNodes", "NODES")
        # fallback=False => new tree all the time
        tree = nodes.new_tree(  # type: ignore
            name=f"MN_{self.name}", input_name="Atoms", is_modifier=True, fallback=False
        )
        self.object.modifiers[0].node_group = tree  # type: ignore

    def get_view(self) -> List[tuple]:
        """
        Get the 3D bounding box of the entity object

        """
        return blender_utils.get_bounding_box(self.object)



import bpy


def check_online_access_for_ui(layout: bpy.types.UILayout) -> bpy.types.UILayout:
    """
    Disable UI without Online Access

    Checks for the online access permissions, and adds a warning and disables following
    UI elements if it fails the check. Returns the UILayout that will have .enabled flag
    set to False, disabling all subsequent uses of the layout.

    Args:
        layout (bpy.types.UILayout): The UILayout element to add the warning and potentially
        disable.

    Returns:
        bpy.types.UILayout: The altered UILayout element, for use in downstream UI
        components.
    """
    if not bpy.app.online_access:
        layout.label(
            text="Online access disabled. Change in Blender's system preferences.",
            icon="ERROR",
        )
        op = layout.operator("wm.url_open", text="Online Access Docs", icon="URL")
        op.url = "https://docs.blender.org/manual/en/dev/editors/preferences/system.html#bpy-types-preferencessystem-use-online-access"
        layout = layout.column()
        layout.alert = True
        layout.enabled = False

    return layout



import numpy as np
from syrupy.extensions.amber import AmberSnapshotExtension


# we create a custom snapshot comparison class, which can handle numpy arrays
# and compare them properly. The class will serialize the numpy arrays into lists
# and when comparing them, reads the list back into a numpy array for comparison
# it checks for 'isclose' for floats and otherwise looks for absolute comparison
class NumpySnapshotExtension(AmberSnapshotExtension):
    def __init__(self):
        super().__init__()
        self.custom_suffix: str | None = None

    def serialize(self, data, cutoff=1000, **kwargs):
        if isinstance(data, np.ndarray):
            shape = data.shape
            if len(shape) == 1:
                if len(data) > cutoff:
                    data = data[:cutoff]
                else:
                    data = data[: int(cutoff / 10),]

            return np.array2string(
                data, precision=1, threshold=2e3, floatmode="maxprec_equal"
            )
        return super().serialize(data, **kwargs)



import os
import shutil
import tempfile
from contextlib import ExitStack
from enum import Enum
from pathlib import Path
from typing import Literal, Tuple
import bpy
from tqdm.auto import tqdm
from .. import assets
from ..assets.template import list_templates
from ..blender import utils as blender_utils
from ..entities.base import MolecularEntity
from ..scene.compositor import setup_compositor
from ..session import get_session
from ..ui import addon
from ..utils import suppress_stdout, temp_override_properties
from .camera import Camera, Viewpoints
from .engines import EEVEE, Cycles

try:
    from IPython.display import Image, Video, display
except ImportError:
    Image = None
    Video = None
    display = None

IS_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"
IS_SELF_HOSTED = os.getenv("RUNNER_ENVIRONMENT") == "self-hosted"


class ViewTransform(Enum):
    STANDARD = "Standard"
    KHRONOS = "Khronos PBR Neutral"
    AGX = "AgX"
    FILMIC = "Filmic"
    FILMIC_LOG = "Filmic Log"
    FALSE_COLOR = "False Color"
    RAW = "Raw"


_view_transform = Literal[
    ViewTransform.STANDARD.value,
    ViewTransform.KHRONOS.value,
    ViewTransform.AGX.value,
    ViewTransform.FILMIC.value,
    ViewTransform.FILMIC_LOG.value,
    ViewTransform.FALSE_COLOR.value,
    ViewTransform.RAW.value,
]


class Canvas:
    """
    High-level render controller for Blender scenes.

    Canvas configures the active Blender scene for Molecular Nodes renders
    (engine, resolution, transparency, color management), exposes convenient
    properties for common render settings, and provides helpers to frame
    objects/views and render stills or animations.

    Parameters
    ----------
    engine : EEVEE | Cycles | str, default "EEVEE"
        Render engine to use. Accepts an instance of ``mn.scene.EEVEE`` or
        ``mn.scene.Cycles``, or a case-insensitive string: ``"EEVEE"`` or
        ``"CYCLES"``.
    resolution : tuple[int, int], default (1280, 720)
        Output resolution in pixels as ``(width, height)``.
    transparent : bool, default False
        When ``True``, renders use a transparent film (alpha background).
    template : pathlib.Path | str | None, default "Molecular Nodes"
        Scene template to initialize. If a string is provided it can be either
        the name of an installed Blender app template (e.g. ``"Molecular Nodes"``),
        or a path to a ``.blend`` file. If ``None``, the Blender default startup
        file is used.

    Attributes
    ----------
    scene : bpy.types.Scene
        The active Blender scene.
    camera : molecularnodes.scene.camera.Camera
        Convenience camera controller bound to the active scene camera.
    engine : EEVEE | Cycles
        The configured render engine object.
    resolution : tuple[int, int]
        Current render resolution in pixels.
    transparent : bool
        Whether the film background is transparent.
    fps : float
        Frames per second for animation output.
    frame_start : int
        Start frame of the scene range.
    frame_end : int
        End frame of the scene range.
    background : tuple[float, float, float, float]
        World background color as RGBA in the range [0, 1].
    view_transform : {"Standard", "Khronos PBR Neutral", "AgX", "Filmic", "Filmic Log", "False Color", "Raw"}
        Active view transform for color management.

    Examples
    --------
    Create a canvas and render a snapshot with a transparent background::

        import molecularnodes as mn
        cv = mn.Canvas(engine="CYCLES", resolution=(800, 800), transparent=True)
        cv.snapshot("frame.png")

    See Also
    --------
    molecularnodes.scene.engines.EEVEE : Render engine configuration for EEVEE.
    molecularnodes.scene.engines.Cycles : Render engine configuration for Cycles.
    molecularnodes.scene.camera.Camera : Camera controller used by Canvas.
    """

    def __init__(
        self,
        engine: EEVEE | Cycles | str = "EEVEE",
        resolution=(1280, 720),
        transparent: bool = False,
        template: Path | str | None = "Molecular Nodes",
    ) -> None:
        """
        Initialize a Canvas and prepare the Blender scene.

        Parameters
        ----------
        engine : EEVEE | Cycles | str, default "EEVEE"
            Render engine configuration or engine name.
        resolution : tuple[int, int], default (1280, 720)
            Output resolution in pixels as ``(width, height)``.
        transparent : bool, default False
            Enable a transparent film (alpha background) when ``True``.
        template : pathlib.Path | str | None, default "Molecular Nodes"
            Scene template name or path to a ``.blend`` file. Use ``None`` for
            Blender's default startup scene.
        """
        addon.register()
        assets.install()
        if template:
            self.scene_reset(template=template)
        self.engine = engine
        self.resolution = resolution
        self.camera = Camera()
        self.transparent = transparent
        setup_compositor(self.scene)

    @property
    def scene(self) -> bpy.types.Scene:
        """
        Get the active Blender scene.

        Returns
        -------
        bpy.types.Scene
            The current context scene.
        """
        return bpy.context.scene

    @property
    def resolution(self) -> tuple[int, int]:
        """
        Get the render resolution.

        Returns
        -------
        tuple[int, int]
            A tuple containing the x and y resolution values.
        """
        x: int = self.scene.render.resolution_x
        y: int = self.scene.render.resolution_y
        return (x, y)

    @resolution.setter
    def resolution(self, value: tuple[int, int]) -> None:
        """
        Set the render resolution.

        Parameters
        ----------
        value : tuple[int, int]
            A tuple containing the x and y resolution values.
        """
        self.scene.render.resolution_x = value[0]
        self.scene.render.resolution_y = value[1]

    @property
    def engine(self) -> Cycles | EEVEE:
        """
        Get the configured render engine.

        Returns
        -------
        EEVEE | Cycles
            The active engine configuration object.
        """
        return self._engine

    @engine.setter
    def engine(self, value: Cycles | EEVEE | str) -> None:
        """
        Set the render engine.

        Parameters
        ----------
        value : EEVEE | Cycles | str
            Either an engine configuration instance or a case-insensitive
            string: ``"EEVEE"`` or ``"CYCLES"``.

        Raises
        ------
        ValueError
            If an unsupported string is provided.
        ValueError
            If the value is neither a valid string nor supported engine type.
        """
        if isinstance(value, Cycles) or isinstance(value, EEVEE):
            self._engine = value
        elif isinstance(value, str):
            if value.upper() == "CYCLES":
                self._engine = Cycles()
            elif value.upper() in [
                "EEVEE",
                "BLENDER_EEVEE_NEXT",
                "BLENDER_EEVEE",
            ]:
                self._engine = EEVEE()
            else:
                raise ValueError("String does not match either 'EEVEE' or 'CYCLES'")
        else:
            raise ValueError(
                "Must be either a string selecting the render engine or a dataclass mn.scene.Cycles()"
            )

        self._engine._enable_engine()

    @property
    def fps(self) -> float:
        """
        Get the frames per second setting.

        Returns
        -------
        float
            The current FPS value.
        """
        return self.scene.render.fps

    @fps.setter
    def fps(self, value: float) -> None:
        """
        Set the frames per second.

        Parameters
        ----------
        value : float
            The FPS value to set.
        """
        self.scene.render.fps = value

    @property
    def frame_start(self) -> int:
        """
        Get the start frame of the animation.

        Returns
        -------
        int
            The start frame number.
        """
        return self.scene.frame_start

    @frame_start.setter
    def frame_start(self, value: int) -> None:
        """
        Set the start frame of the animation.

        Parameters
        ----------
        value : int
            The start frame number to set.
        """
        self.scene.frame_start = value

    @property
    def frame_end(self) -> int:
        """
        Get the end frame of the animation.

        Returns
        -------
        int
            The end frame number.
        """
        return self.scene.frame_end

    @frame_end.setter
    def frame_end(self, value: int) -> None:
        """
        Set the end frame of the animation.

        Parameters
        ----------
        value : int
            The end frame number to set.
        """
        self.scene.frame_end = value

    @property
    def transparent(self) -> bool:
        """
        Get the transparency setting for rendering.

        Returns
        -------
        bool
            True if transparency is enabled, False otherwise.
        """
        return self.scene.render.film_transparent

    @transparent.setter
    def transparent(self, value: bool) -> None:
        """
        Set the transparency setting for rendering.

        Parameters
        ----------
        value : bool
            True to enable transparency, False to disable.
        """
        self.scene.render.film_transparent = value

    @property
    def background(self) -> Tuple[float, float, float, float]:
        """
        Get the world background color.

        Returns
        -------
        tuple[float, float, float, float]
            RGBA values in the range [0, 1].
        """
        return (
            self.scene.world.node_tree.nodes["MN_world_shader"].inputs[3].default_value
        )

    @background.setter
    def background(self, value: Tuple[float, float, float, float]) -> None:
        """
        Set the world background color.

        Parameters
        ----------
        value : tuple[float, float, float, float]
            RGBA values in the range [0, 1].
        """
        self.scene.world.node_tree.nodes["MN_world_shader"].inputs[
            3
        ].default_value = value

    # @property
    # def hdri_strength

    @property
    def view_transform(self) -> _view_transform:
        """
        Get the current view transform setting.

        Returns
        -------
        str
            The current view transform value.
        """
        return self.scene.view_settings.view_transform

    @view_transform.setter
    def view_transform(self, value: _view_transform | ViewTransform | str) -> None:
        """
        Set the view transform setting.

        Parameters
        ----------
        value : str | ViewTransform
            The view transform value to set. Accepts enum, full name, or lowercase/shortened name.
        """
        # Normalize input
        if isinstance(value, ViewTransform):
            vt_value = value.value
        elif isinstance(value, str):
            value_lower = value.strip().lower()
            # Try to match full name (case-insensitive)
            for vt in ViewTransform:
                if value_lower == vt.value.lower():
                    vt_value = vt.value
                    break
            else:
                # Try to match shortened name (e.g., "standard", "agx", "filmic", etc.)
                for vt in ViewTransform:
                    if value_lower == vt.name.lower():
                        vt_value = vt.value
                        break
                else:
                    raise ValueError(
                        f"Invalid view transform '{value}'. Must be one of {[vt.value for vt in ViewTransform]} or a valid enum name."
                    )
        else:
            raise TypeError("view_transform must be a str or ViewTransform enum.")

        self.scene.view_settings.view_transform = vt_value

    def frame_object(
        self, obj: bpy.types.Object | MolecularEntity, viewpoint: Viewpoints = None
    ) -> None:
        """
        Frame an object or MolecularEntity in the active camera view.

        Parameters
        ----------
        obj : bpy.types.Object | MolecularEntity
            Blender object or Molecular Nodes entity to frame.
        viewpoint : Viewpoints | str, optional
            Viewing direction along a principal axis. One of
            {"default", "front", "back", "top", "bottom", "left", "right"}.

        """
        if isinstance(obj, MolecularEntity):
            obj = obj.object
        # set the camera viewpoint if specified
        if viewpoint is not None:
            self.camera.set_viewpoint(viewpoint)
        # set the camera to look at the object
        blender_utils.look_at_object(obj)

    def frame_view(
        self, view: list[tuple] | MolecularEntity, viewpoint: Viewpoints | None = None
    ) -> None:
        """
        Frame one or more views of Molecular entities.
        Multiple views can be combined using ``+`` before passing the result.

        Parameters
        ----------
        view : list[tuple] | MolecularEntity
            A bounding box represented by 8 three-dimensional vertices
            ``[(x, y, z), ...]`` or an entity from which a view is derived.
        viewpoint : Viewpoints | str, optional
            Viewing direction along a principal axis. One of
            {"default", "front", "back", "top", "bottom", "left", "right"}.

        """
        if isinstance(view, MolecularEntity):
            view_tuple = view.get_view()
        else:
            view_tuple = view
        # set the camera viewpoint if specified
        if viewpoint is not None:
            self.camera.set_viewpoint(viewpoint)
        # set the camera to look at the bounding box of the view
        blender_utils.look_at_bbox(view_tuple)

    def clear(self) -> None:
        """
        Clear all Molecular Nodes entities from the scene.

        Notes
        -----
        This does not modify lighting, world, or render settings.
        """
        session = get_session()
        # Iterate over a list copy to avoid modifying the dict during iteration
        for entity in list(session.entities.values()):
            session.remove(entity.uuid)

    def scene_reset(
        self,
        template: Path | str | None = "Molecular Nodes",
        engine: Cycles | EEVEE | str = "EEVEE",
    ) -> None:
        """
        Reset the scene from a template or startup file.

        Parameters
        ----------
        template : pathlib.Path | str | None, default "Molecular Nodes"
            Name of an installed Blender app template, a path to a ``.blend``
            file, or ``None`` to use Blender's default startup file.
        engine : EEVEE | Cycles | str, default "EEVEE"
            Render engine to configure after loading the template.

        Raises
        ------
        ValueError
            If ``template`` is not ``None``, not a valid ``.blend`` file path,
            and not a known app template name.
        """
        if template is None:
            bpy.ops.wm.read_homefile(app_template="")
        else:
            file = Path(template) if isinstance(template, str) else template
            if file.is_file() and file.suffix == ".blend":
                self.load(file)
            elif isinstance(template, str):
                if template in list_templates():
                    bpy.ops.wm.read_homefile(app_template=template)
                else:
                    raise ValueError(
                        f"Template '{template}' is not a valid .blend file or app template name."
                    )

            else:
                raise ValueError(
                    f"Template '{template}' is not a valid .blend file or app template name."
                )

        if engine:
            self.engine = engine

    def load(self, path: str | Path) -> None:
        """
        Load a .blend file replacing the current scene.

        Parameters
        ----------
        path : str | Path
            The file path to the .blend file to load.
        """
        file_path = Path(path) if isinstance(path, str) else path
        if not file_path.is_file() or file_path.suffix != ".blend":
            raise ValueError(f"File '{path}' is not a valid .blend file.")
        bpy.ops.wm.open_mainfile(filepath=str(file_path.resolve()))

    def snapshot(
        self,
        path: str | Path | None = None,
        frame: int | None = None,
        file_format: str = "PNG",
    ) -> None:
        """
        Render an image of the current scene.

        Parameters
        ----------
        path : str | Path | None, optional
            File path to write the rendered image to.

        frame : int, optional
            Frame number of scene to render. When not specified,
            current scene's current_frame is used

        file_format : str, optional
            File format of the rendered image.

        """
        scene = self.scene
        render_settings = scene.render
        image_settings = render_settings.image_settings
        render_frame = scene.frame_current if frame is None else frame
        # temporary properties to override
        override_props = [
            (render_settings, "use_file_extension", True),
            (scene, "frame_current", render_frame),
        ]
        if bpy.app.version >= (5, 0, 0):
            override_props.append((image_settings, "media_type", "IMAGE"))
        override_props.append((image_settings, "file_format", file_format))
        with ExitStack() as stack:
            # set the use_file_extension to auto generate file extension
            # set the file_format to the specified one
            temp_override_properties(stack, override_props)
            # create temporary file with the file_format extension
            tmp_file = stack.enter_context(
                tempfile.NamedTemporaryFile(suffix=render_settings.file_extension)
            )
            # set the filepath to the temporary file
            temp_override_properties(
                stack,
                [
                    (render_settings, "filepath", tmp_file.name),
                ],
            )
            # set the frame number - only frame_set will update animation data
            # scene.current_frame will only update the timeline
            scene.frame_set(render_frame)
            # suppress stdout output generated by render process
            with suppress_stdout():
                # render the image
                bpy.ops.render.render(write_still=True, animation=False)
            if path:
                # save to file if path specified
                shutil.copy(tmp_file.name, path)
            elif display and Image:
                # only display in notebook if path not specified
                # and in notebook context
                display(Image(tmp_file.name))

    def animation(
        self,
        path: str | Path | None = None,
        frame_start: int | None = None,
        frame_end: int | None = None,
        render_scale: int = 100,
    ) -> None:
        """
        Render an animation of the current scene.

        Parameters
        ----------
        path : str | Path | None, optional
            File path to write the rendered animation to.

        frame_start : int, optional
            Start frame of the animation. When not specified, current scene's
            start frame is used

        frame_end : int, optional
            End frame of the animation. When not specified, current scene's
            end frame is used

        render_scale : int, optional
            Scale of the rendered animation frames with respect to the resolution.

        """
        # determine frame range
        start = self.frame_start if frame_start is None else frame_start
        end = self.frame_end if frame_end is None else frame_end
        if end < start:
            raise ValueError(f"End frame {end} cannot be less than start frame {start}")
        n = len(str(end))
        frame_range = range(start, end + 1)

        scene = self.scene
        render_settings = scene.render
        image_settings = render_settings.image_settings
        # temporary properties to override
        override_props = [
            (render_settings, "use_lock_interface", True),
            (render_settings, "resolution_percentage", render_scale),
            (render_settings, "filepath", ""),
            (scene, "frame_current", start),
        ]
        if bpy.app.version >= (5, 0, 0):
            override_props.append((image_settings, "media_type", "IMAGE"))
        override_props.append((image_settings, "file_format", "PNG"))
        # create a temporary directory
        with tempfile.TemporaryDirectory() as tmp_dir:
            # render individual frames
            with ExitStack() as stack:
                temp_override_properties(stack, override_props)
                it = tqdm(frame_range, desc="Rendering frames")
                # render individual frames
                for frame in it:
                    render_image = os.path.join(tmp_dir, str(frame).zfill(n)) + ".png"
                    # set the output file
                    scene.render.filepath = render_image
                    # set the frame - only frame_set updates animation data
                    scene.frame_set(frame)
                    with suppress_stdout():
                        # render each frame image
                        bpy.ops.render.render(write_still=True)
                it.close()

            # add to video sequence editor
            sequence_editor = scene.sequence_editor_create()
            strips = sequence_editor.strips  # .sequences is deprecated
            image_strip = None
            for frame in frame_range:
                render_image = os.path.join(tmp_dir, str(frame).zfill(n)) + ".png"
                if image_strip is None:
                    image_strip = strips.new_image("name", render_image, 1, start)
                else:
                    image_strip.elements.append(os.path.basename(render_image))

            # render animation
            video_file = os.path.join(tmp_dir, "animation.mp4")
            # temporary properties to override
            override_props = [
                (render_settings, "use_lock_interface", True),
                (render_settings, "resolution_percentage", render_scale),
                (render_settings, "filepath", ""),
                (scene, "frame_current", start),
                (scene, "frame_start", start),
                (scene, "frame_end", end),
            ]
            if bpy.app.version >= (5, 0, 0):
                override_props.append((image_settings, "media_type", "VIDEO"))
            override_props.append((image_settings, "file_format", "FFMPEG"))
            override_props.append((render_settings.ffmpeg, "format", "MPEG4"))
            with ExitStack() as stack:
                temp_override_properties(stack, override_props)
                scene.render.filepath = video_file
                it = tqdm(range(0, 1), desc="Generating video")
                for i in it:
                    with suppress_stdout():
                        # render the animation of sequence editor images
                        bpy.ops.render.render(animation=True)
                it.close()

            # clear the video sequence editor
            strips.remove(image_strip)

            if path:
                # save to file if path specified
                shutil.copy(video_file, path)
            elif display and Video:
                # only display in notebook if path not specified
                # and in notebook context
                display(Video(video_file, embed=True))



import bpy
from databpy.object import LinkedObjectError
from ..entities import StreamingTrajectory, density, trajectory
from ..entities.base import EntityType
from ..nodes import nodes
from ..nodes.geometry import get_final_style_nodes
from ..session import get_session
from .pref import addon_preferences
from .utils import check_online_access_for_ui


def panel_wwpdb(layout, scene):
    layout.label(text="Download from PDB", icon="IMPORT")
    layout.separator()

    layout = check_online_access_for_ui(layout)

    row_import = layout.row().split(factor=0.5)
    row_import.prop(scene.mn, "import_code_pdb")
    row = row_import.split(factor=0.3)
    row.prop(scene.mn, "import_format_fetch", text="")
    op = row.operator("mn.import_fetch")
    op.code = scene.mn.import_code_pdb  # type: ignore
    op.database = "wwpdb"  # type: ignore
    op.file_format = scene.mn.import_format_fetch  # type: ignore
    op.node_setup = scene.mn.import_node_setup  # type: ignore
    op.remove_solvent = scene.mn.import_remove_solvent  # type: ignore
    op.assembly = scene.mn.import_build_assembly  # type: ignore
    op.style = scene.mn.import_style  # type: ignore
    op.centre = scene.mn.import_centre  # type: ignore
    op.centre_type = scene.mn.import_centre_type  # type: ignore
    prefs = addon_preferences()
    if prefs is not None:
        op.cache_dir = str(prefs.cache_dir)  # type: ignore
    else:
        op.cache_dir = str(bpy.app.tempdir)  # type: ignore
    layout.separator(factor=0.4)

    layout.separator()

    layout.label(text="Options", icon="MODIFIER")
    options = layout.column(align=True)

    row = options.row()
    row.prop(scene.mn, "import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.mn.import_node_setup

    row_centre = options.row()
    row_centre.prop(scene.mn, "import_centre", icon_value=0)
    col_centre = row_centre.column()
    col_centre.prop(scene.mn, "import_centre_type", text="")
    col_centre.enabled = scene.mn.import_centre
    options.separator()

    grid = options.grid_flow()
    grid.prop(scene.mn, "import_build_assembly")
    grid.prop(scene.mn, "import_remove_solvent")
    grid.prop(scene.mn, "import_del_hydrogen")


def panel_alphafold(layout, scene):
    layout.label(text="Download from the AlphaFold DataBase", icon="IMPORT")
    layout.separator()

    layout = check_online_access_for_ui(layout)

    row_import = layout.row().split(factor=0.5)
    row_import.prop(scene.mn, "import_code_alphafold")
    download = row_import.split(factor=0.3)
    download.prop(scene.mn, "import_format_fetch", text="")
    op = download.operator("mn.import_fetch")
    op.code = scene.mn.import_code_alphafold  # type: ignore
    op.database = "alphafold"  # type: ignore
    op.file_format = scene.mn.import_format_fetch  # type: ignore
    op.node_setup = scene.mn.import_node_setup  # type: ignore
    op.assembly = scene.mn.import_build_assembly  # type: ignore
    op.style = scene.mn.import_style  # type: ignore
    op.centre = scene.mn.import_centre  # type: ignore
    op.centre_type = scene.mn.import_centre_type  # type: ignore
    prefs = addon_preferences()
    if prefs is not None:
        op.cache_dir = str(prefs.cache_dir)  # type: ignore
    else:
        op.cache_dir = str(bpy.app.tempdir)  # type: ignore

    layout.separator(factor=0.4)

    row = layout.row().split(factor=0.3)
    layout.separator()

    layout.label(text="Options", icon="MODIFIER")
    options = layout.column(align=True)

    row = options.row()
    row.prop(scene.mn, "import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.mn.import_node_setup

    row_centre = options.row()
    row_centre.prop(scene.mn, "import_centre", icon_value=0)
    col_centre = row_centre.column()
    col_centre.prop(scene.mn, "import_centre_type", text="")
    col_centre.enabled = scene.mn.import_centre
    options.separator()


# operator that calls the function to import the structure from a local file


def panel_local(layout, scene):
    layout.label(text="Load a Local File", icon="FILE_TICK")
    layout.separator()

    row = layout.row()
    row.prop(scene.mn, "import_local_path")
    op = row.operator("mn.import_local")
    op.filepath = scene.mn.import_local_path
    op.node_setup = scene.mn.import_node_setup
    op.assembly = scene.mn.import_build_assembly
    op.style = scene.mn.import_style
    op.centre = scene.mn.import_centre
    op.remove_solvent = scene.mn.import_remove_solvent
    op.centre_type = scene.mn.import_centre_type
    layout.separator()

    layout.label(text="Options", icon="MODIFIER")
    options = layout.column(align=True)

    row = options.row()
    row.prop(scene.mn, "import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.mn.import_node_setup

    row_centre = options.row()

    row_centre.prop(scene.mn, "import_centre", icon_value=0)
    # row_centre.prop()
    col_centre = row_centre.column()
    col_centre.prop(scene.mn, "import_centre_type", text="")
    col_centre.enabled = scene.mn.import_centre
    options.separator()

    grid = options.grid_flow()
    grid.prop(scene.mn, "import_build_assembly")
    grid.prop(scene.mn, "import_remove_solvent", icon_value=0)
    grid.prop(scene.mn, "import_del_hydrogen", icon_value=0)


def panel_starfile(layout, scene):
    layout.label(text="Load Star File", icon="FILE_TICK")
    layout.separator()
    row_import = layout.row()
    row_import.prop(scene.mn, "import_star_file_path")
    op = row_import.operator("mn.import_star_file")
    op.filepath = scene.mn.import_star_file_path
    op.node_setup = scene.mn.import_node_setup


def panel_cellpack(layout, scene):
    layout.label(text="Load CellPack Model", icon="FILE_TICK")
    layout.separator()
    row = layout.row()
    row.prop(scene.mn, "import_cell_pack_path")
    op = row.operator("mn.import_cell_pack")
    op.filepath = scene.mn.import_cell_pack_path
    op.node_setup = scene.mn.import_node_setup


def panel_density(layout, scene):
    layout.label(text="Load Density Grids", icon="FILE_TICK")
    layout.separator()

    row = layout.row()
    row.prop(scene.mn, "import_density")
    row.operator("mn.import_density")

    layout.separator()
    col = layout.column()
    col.alignment = "LEFT"
    col.scale_y = 0.5
    label = f"\
    An intermediate file will be created: {scene.mn.import_density}.vdb\
    Please do not delete this file or the volume will not render.\
    Move the original .map file to change this location.\
    "
    for line in label.strip().split("    "):
        col.label(text=line)

    layout.separator()
    layout.label(text="Options", icon="MODIFIER")

    layout.prop(scene.mn, "import_density_invert")
    layout.prop(scene.mn, "import_density_center")
    layout.prop(scene.mn, "import_density_overwrite")
    row = layout.row()
    row.prop(scene.mn, "import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_density_style")
    col.enabled = scene.mn.import_node_setup


def panel_trajectory(layout, scene):
    layout.label(text="Load MD Trajectories", icon="FILE_TICK")
    layout.separator()
    col = layout.column(align=True)
    row_import = col.row()
    row_import.prop(scene.mn, "import_md_name")
    op = row_import.operator("mn.import_trajectory", text="Load")
    op.topology = scene.mn.import_md_topology
    op.trajectory = scene.mn.import_md_trajectory
    op.name = scene.mn.import_md_name
    op.style = scene.mn.import_style
    op.setup_nodes = scene.mn.import_node_setup
    col.separator()
    col.prop(scene.mn, "import_md_topology")
    col.prop(scene.mn, "import_md_trajectory")

    layout.separator()
    layout.label(text="Options", icon="MODIFIER")
    row = layout.row()
    row.prop(scene.mn, "import_node_setup", text="")
    col = row.column()
    col.prop(scene.mn, "import_style")
    col.enabled = scene.mn.import_node_setup


def panel_oxdna(layout: bpy.types.UILayout, scene: bpy.types.Scene) -> None:
    """
    Create the panel layout for oxDNA import.

    Parameters
    ----------
    layout : bpy.types.UILayout
        Layout to add elements to
    scene : bpy.types.Scene
        Current scene
    """
    layout.label(text="Load oxDNA File", icon="FILE_TICK")
    layout.separator()
    row = layout.row()
    row.prop(scene.mn, "import_oxdna_name")
    op = row.operator("mn.import_oxdna")
    op.name = scene.mn.import_oxdna_name
    op.topology = scene.mn.import_oxdna_topology
    op.trajectory = scene.mn.import_oxdna_trajectory
    col = layout.column(align=True)
    col.prop(scene.mn, "import_oxdna_topology")
    col.prop(scene.mn, "import_oxdna_trajectory")


chosen_panel = {
    "pdb": panel_wwpdb,
    "local": panel_local,
    "alphafold": panel_alphafold,
    "star": panel_starfile,
    "md": panel_trajectory,
    "density": panel_density,
    "cellpack": panel_cellpack,
    "dna": panel_oxdna,
}


def pt_object_context(self, context):
    return None


def is_style_node(context):
    node = context.space_data.edit_tree.nodes.active
    return node.name.startswith("Style")


def change_style_node_menu(self, context):
    layout = self.layout
    node = context.active_node

    # return early if not a node group
    if not hasattr(node, "node_tree"):
        return None

    # return early if the node group isn't one of the ones we want to swap easily
    prefix = node.node_tree.name.split(" ")[0].lower()
    if prefix not in ["color", "select", "is", "style", "topology", "animate"]:
        return None

    layout.label(text="Molecular Nodes", icon="MOD_PARTICLES")

    row = layout.row()
    op = row.operator_menu_enum("mn.node_swap", "node_items", text="Change Node")
    op.node_description = "The topology nodes"

    layout.separator()


def panel_import(layout, context):
    scene = context.scene
    selection = scene.mn.panel_import_type
    layout.prop(scene.mn, "panel_import_type")

    col = layout.column()
    chosen_panel[selection](col, scene)


def ui_from_node(
    layout: bpy.types.UILayout,
    node: bpy.types.GeometryNodeGroup,
    context: bpy.types.Context,
):
    """
    Generate the UI for a particular node, which displays the relevant node inputs
    for user control in a panel, rather than through the node editor.
    """
    col = layout.column(align=True)
    ntree = context.active_object.modifiers["MolecularNodes"].node_group

    tree = node.node_tree.interface.items_tree

    for item in tree.values():
        if item.item_type == "PANEL":
            col.label(text=item.name)
        elif item.name == "Selection":
            continue
        else:
            if item.in_out != "INPUT":
                continue
            if item.socket_type == "NodeSocketGeometry":
                continue
            col.template_node_view(ntree, node, node.inputs[item.identifier])


def panel_md_properties(layout, context):
    obj = context.active_object
    session = get_session()
    traj: trajectory.Trajectory = session.match(obj)
    traj_is_linked = bool(traj)
    if traj is not None and not isinstance(traj, trajectory.Trajectory):
        raise TypeError(f"Expected a trajectory, got {type(traj)}")

    col = layout.column()
    col.enabled = False
    if not traj_is_linked:
        col.enabled = True
        col.label(text="Object not linked to a trajectory, please reload one")
        col.prop(obj.mn, "filepath_topology")
        col.prop(obj.mn, "filepath_trajectory")
        col.operator("mn.reload_trajectory")
        return None

    layout.label(text="Trajectory Playback", icon="OPTIONS")
    is_streaming = isinstance(traj, StreamingTrajectory)

    if is_streaming:
        label = "Streaming trajectory; cannot alter playback"
    else:
        label = "Trajectory has {} {}".format(
            obj.mn.n_frames, "frames" if obj.mn.n_frames else "frame"
        )

    layout.label(text=label)
    playback = layout.column()
    playback.enabled = not is_streaming
    row = playback.row()

    col = row.column()
    if obj.mn.update_with_scene:
        col.prop(obj.mn, "frame_hidden")
    else:
        col.prop(obj.mn, "frame")
    col.enabled = not obj.mn.update_with_scene
    row.prop(obj.mn, "update_with_scene")
    row = playback.row()
    col = row.column()
    col.enabled = obj.mn.update_with_scene
    col.prop(obj.mn, "average")
    col.prop(obj.mn, "subframes")
    col.prop(obj.mn, "offset")
    col = row.column()
    col.enabled = obj.mn.update_with_scene

    # only enable this as an option if the universe is orthothombic
    row = col.row()
    row.prop(obj.mn, "correct_periodic")
    row.enabled = traj._is_orthorhombic
    col.prop(obj.mn, "interpolate")

    layout.label(text="Selections", icon="RESTRICT_SELECT_OFF")
    row = layout.row()
    row = row.split(factor=0.9)
    row.template_list(
        "MN_UL_TrajectorySelectionListUI",
        "A list",
        obj,
        "mn_trajectory_selections",
        obj.mn,
        "trajectory_selection_index",
        rows=3,
    )
    col = row.column()
    col.operator("mn.trajectory_selection_add", icon="ADD", text="")
    col.operator("mn.trajectory_selection_remove", icon="REMOVE", text="")
    if obj.mn_trajectory_selections:
        item = obj.mn_trajectory_selections[obj.mn.trajectory_selection_index]

        col = layout.column(align=False)
        row = col.row()
        col.prop(item, "string")

        # disable editing for immutable selections
        # disable modifying updating and periodic
        if item.from_atomgroup:
            col.enabled = False

        if item.message != "":
            box = col.box()
            box.label(text="Invalid Selection", icon="ERROR")
            box.label(text=item.message)
            box.alert = True
            op = box.operator("wm.url_open", text="Selection Langauge Docs", icon="URL")
            op.url = (
                "https://docs.mdanalysis.org/stable/documentation_pages/selections.html"
            )


def panel_object(layout, context):
    object = context.active_object
    if object is None:
        # When an object is deleted, context.ative_object is None
        return
    row = layout.row()
    row.prop(object.mn, "entity_type")
    row.enabled = False
    try:
        mol_type = object.mn.entity_type
    except AttributeError:
        return None
    if mol_type == "None":
        layout.label(text="No MN object selected")
        return None
    if mol_type.startswith("md"):
        panel_md_properties(layout, context)
    if mol_type == "ensemble-star":
        layout.label(text="Ensemble")
        box = layout.box()
        ui_from_node(box, nodes.get_star_node(object), context=context)
        return None


def item_ui(layout, item):
    row = layout.row()
    row.label(text=item.name)
    col = row.column()
    op = col.operator("mn.session_create_object")
    op.uuid = item.uuid
    col.enabled = item.object is None

    op = row.operator("mn.session_remove_item", text="", icon="CANCEL")
    op.uuid = item.uuid

    if item.object is not None:
        row = layout.row()
        row.label(text=f"Object: {item.object.name}", icon="OUTLINER_OB_MESH")


def panel_session(layout, context):
    session = get_session(context)
    row = layout.row()
    row.label(text="Loaded items in the session")

    layout.label(text="Molecules")
    box = layout.box()
    for mol in session.molecules.values():
        item_ui(box, mol)

    layout.label(text="Universes")
    box = layout.box()
    for uni in session.trajectories.values():
        item_ui(box, uni)

    layout.label(text="Ensembles")
    box = layout.box()
    for ens in session.ensembles.values():
        item_ui(box, ens)


def panel_scene(layout, context):
    scene = context.scene

    cam = bpy.data.cameras[bpy.data.scenes["Scene"].camera.name]
    world_shader = bpy.data.worlds["World Shader"].node_tree.nodes["MN_world_shader"]
    grid = layout.grid_flow()
    col = grid.column()
    col.label(text="World Settings")
    world = col.box()
    world.prop(bpy.data.scenes["Scene"].render, "engine")
    if scene.render.engine == "CYCLES":
        world.prop(bpy.data.scenes["Scene"].cycles, "samples")
    else:
        world.prop(bpy.data.scenes["Scene"].eevee, "taa_render_samples")
    world.label(text="Background")
    world.prop(world_shader.inputs[1], "default_value", text="HDRI Strength")
    row = world.row()
    row.prop(scene.render, "film_transparent")
    row.prop(world_shader.inputs[2], "default_value", text="Background")

    col = grid.column()
    col.label(text="Camera Settings")
    camera = col.box()
    camera.prop(cam, "lens")
    col = camera.column(align=True)
    row = col.row(align=True)
    row.prop(bpy.data.scenes["Scene"].render, "resolution_x", text="X")
    row.prop(bpy.data.scenes["Scene"].render, "resolution_y", text="Y")
    row = camera.grid_flow()
    row.prop(cam.dof, "use_dof")
    row.prop(bpy.data.scenes["Scene"].render, "use_motion_blur")
    focus = camera.column()
    focus.enabled = cam.dof.use_dof
    focus.prop(cam.dof, "focus_object")
    distance = focus.row()
    distance.enabled = cam.dof.focus_object is None
    distance.prop(cam.dof, "focus_distance")
    focus.prop(cam.dof, "aperture_fstop")


class MN_PT_Scene(bpy.types.Panel):
    bl_label = "Molecular Nodes"
    bl_idname = "MN_PT_panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "scene"
    bl_order = 0
    bl_options = {"HEADER_LAYOUT_EXPAND"}
    bl_ui_units_x = 0

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        row = layout.row()

        row = layout.row(align=True)

        for p in ["import", "object", "session"]:
            row.prop_enum(scene.mn, "panel_selection", p)

        # the possible panel functions to choose between
        which_panel = {
            "import": panel_import,
            "object": panel_object,
            "session": panel_session,
        }
        # call the required panel function with the layout and context
        which_panel[scene.mn.panel_selection](layout, context)


class MN_UL_EntitiesList(bpy.types.UIList):
    """
    UIList of entities in Entities panel (Viewport)
    """

    def draw_item(
        self,
        context,
        layout,
        data,
        item,
        icon,
        active_data,
        active_property,
        index=0,
        flt_flag=0,
    ):
        custom_icon = "WORLD"
        if self.layout_type in {"DEFAULT", "COMPACT"}:
            row = layout.row()
            seqno = f"{index + 1}. "
            split = row.split(factor=0.1)
            col = split.column()
            col.label(text=seqno)
            col = split.column()
            session = context.scene.MNSession
            entity = session.get(item.name)
            col.prop(entity.object, "name", text="", emboss=False)
            # use the object viewport visibility to determine the icon
            # we do not have direct callbacks for raw object visibility changes
            hide_icon = "HIDE_OFF" if not entity.object.hide_get() else "HIDE_ON"
            row.prop(
                item,
                "visible",
                icon_only=True,
                icon=hide_icon,
            )
        elif self.layout_type in {"GRID"}:
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)

    def filter_items(self, context, data, propname):
        items = getattr(data, propname)
        filtered = [self.bitflag_filter_item] * len(items)
        for i, item in enumerate(items):
            try:
                _ = context.scene.MNSession.get(item.name).name
            except (LinkedObjectError, AttributeError):
                filtered[i] &= ~self.bitflag_filter_item
        return filtered, []


class MN_PT_Entities(bpy.types.Panel):
    """
    Panel to list MN Entities in Viewport
    """

    bl_idname = "MN_PT_Entities"
    bl_label = "Entities"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Molecular Nodes"

    def draw(self, context):
        layout = self.layout
        props = context.scene.mn
        row = layout.row()
        row.template_list(
            "MN_UL_EntitiesList",
            "entities_list",
            props,
            "entities",
            props,
            "entities_active_index",
            rows=3,
        )
        col = row.column()
        row = col.row()
        row.operator("mn.session_prune", icon="FILE_REFRESH", text="")
        row = col.row()
        row.operator("mn.session_create_object", icon="ADD", text="")
        row.enabled = False  # TODO: create object or create entity or remove?
        row = col.row()
        op = row.operator("mn.session_remove_item", icon="REMOVE", text="")
        if props.entities_active_index == -1:
            row.enabled = False
        else:
            op.uuid = props.entities[props.entities_active_index].name

        if props.entities_active_index == -1:
            return
        # display entity type of the selected entity
        uuid = props.entities[props.entities_active_index].name
        entity = context.scene.MNSession.get(uuid)
        if not entity:
            return
        row = layout.row()
        row.prop(entity.object.mn, "entity_type")
        row.enabled = False


class MN_PT_trajectory(bpy.types.Panel):
    """
    Panel for trajectory details
    """

    bl_idname = "MN_PT_trajectory"
    bl_label = "Trajectory"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Molecular Nodes"

    @classmethod
    def poll(cls, context):
        """Visible only if entity selected is a trajectory"""
        scene = context.scene
        active_index = scene.mn.entities_active_index
        if active_index == -1:
            return False
        uuid = scene.mn.entities[active_index].name
        try:
            return scene.MNSession.get(uuid).object.mn.entity_type in (
                EntityType.MD.value,
                EntityType.MD_OXDNA.value,
            )
        except (LinkedObjectError, AttributeError):
            return False

    def draw(self, context):
        layout = self.layout
        # To enable the animatate dot next to property in UI
        # layout.use_property_split = True
        # layout.use_property_decorate = True
        scene = context.scene
        active_index = scene.mn.entities_active_index
        uuid = scene.mn.entities[active_index].name
        # Use the object corresponding to the entity
        object = scene.MNSession.get(uuid).object
        props = object.mn
        row = layout.row()
        label = "This trajectory has " + str(props.n_frames) + " frames"
        row.label(text=label)
        row = layout.row()
        row.prop(props, "update_with_scene")
        box = layout.box()
        row = box.row()
        row.prop(props, "frame")
        box.enabled = not props.update_with_scene


class MN_UL_StylesList(bpy.types.UIList):
    """
    UIList of styles for an entity
    """

    seqno = 1

    def draw_item(
        self,
        context,
        layout,
        data,
        item,
        icon,
        active_data,
        active_property,
        index=0,
        flt_flag=0,
    ):
        custom_icon = "WORLD"
        if self.layout_type in {"DEFAULT", "COMPACT"}:
            row = layout.row()
            seqno = f"{self.seqno}. "
            MN_UL_StylesList.seqno += 1
            split = row.split(factor=0.1)
            col = split.column()
            col.label(text=seqno)
            col = split.column()
            col.prop(item, "label", text="", emboss=False)
            if "Visible" in item.inputs:
                input = item.inputs["Visible"]
                hide_icon = "HIDE_OFF" if input.default_value else "HIDE_ON"
                row.prop(input, "default_value", icon_only=True, icon=hide_icon)
        elif self.layout_type in {"GRID"}:
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)

    def filter_items(self, context, data, propname):
        items = getattr(data, propname)
        helper_funcs = bpy.types.UI_UL_list
        filtered = []
        # Filtering by name
        if self.filter_name:
            filtered = helper_funcs.filter_items_by_name(
                self.filter_name,
                self.bitflag_filter_item,
                items,
                "label",
                reverse=False,
            )
        if not filtered:
            filtered = [self.bitflag_filter_item] * len(items)
        style_nodes = get_final_style_nodes(data)
        ordered = [-1] * len(items)
        for i, item in enumerate(items):
            if item in style_nodes:
                ordered[i] = style_nodes.index(item)
            else:
                filtered[i] &= ~self.bitflag_filter_item
        index = len(style_nodes)
        for i, item in enumerate(items):
            if ordered[i] == -1:
                ordered[i] = index
                index += 1
        return filtered, ordered


class MN_PT_Styles(bpy.types.Panel):
    """
    Panel for styles
    """

    bl_idname = "MN_PT_styles"
    bl_label = "Styles"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Molecular Nodes"

    @classmethod
    def poll(cls, context):
        """Visible only if entity selected is a trajectory or molecule"""
        scene = context.scene
        active_index = scene.mn.entities_active_index
        if active_index == -1:
            return False
        uuid = scene.mn.entities[active_index].name
        try:
            return scene.MNSession.get(uuid).object.mn.entity_type in (
                EntityType.MD.value,
                EntityType.MD_STREAMING.value,
                EntityType.MOLECULE.value,
                EntityType.DENSITY.value,
            )
        except (LinkedObjectError, AttributeError):
            return False

    def draw(self, context):
        scene = context.scene
        entities_active_index = scene.mn.entities_active_index
        uuid = scene.mn.entities[entities_active_index].name
        entity = scene.MNSession.get(uuid)
        node_group = entity.node_group
        if node_group is None:
            return
        styles_active_index = entity.object.mn.styles_active_index
        valid_selection = False
        style_nodes = get_final_style_nodes(node_group)
        if 0 <= styles_active_index < len(node_group.nodes):
            if node_group.nodes[styles_active_index] in style_nodes:
                valid_selection = True

        layout = self.layout
        row = layout.row()
        MN_UL_StylesList.seqno = 1
        row.template_list(
            "MN_UL_StylesList",
            "styles_list",
            node_group,
            "nodes",
            entity.object.mn,
            "styles_active_index",
            rows=3,
        )
        if not isinstance(entity, density.Density):
            col = row.column()
            row = col.row()
            op = row.operator("mn.add_style", icon="ADD", text="")
            op.uuid = uuid
            row = col.row()
            op = row.operator("mn.remove_style", icon="REMOVE", text="")
            if valid_selection:
                op.uuid = uuid
                op.style_node_index = styles_active_index
            else:
                row.enabled = False

        if not valid_selection:
            return

        box = layout.box()
        row = box.row()
        style_node = node_group.nodes[styles_active_index]

        panels = {}
        for item in style_node.node_tree.interface.items_tree.values():
            if item.item_type == "PANEL":
                header = None
                if item.parent.name and item.parent.name in panels:
                    panel = panels[item.parent.name]
                    if panel:
                        header, panel = panel.panel(item.name, default_closed=False)
                else:
                    header, panel = box.panel(item.name, default_closed=False)
                if header:
                    header.label(text=item.name)
                panels[item.name] = panel
            elif item.name == "Selection":
                continue
            else:
                if item.in_out != "INPUT":
                    continue
                if item.name in ("Visible"):
                    continue
                input = style_node.inputs[item.identifier]
                if not hasattr(input, "default_value"):
                    continue
                row = None
                if item.parent.name and item.parent.name in panels:
                    panel = panels[item.parent.name]
                    if panel:
                        row = panel.row()
                else:
                    row = box.row()
                if row:
                    row.prop(data=input, property="default_value", text=input.name)
        row = box.row()


class MN_UL_AnnotationsList(bpy.types.UIList):
    """
    UIList of annotations for an entity
    """

    def draw_item(
        self,
        context,
        layout,
        data,
        item,
        icon,
        active_data,
        active_property,
        index=0,
        flt_flag=0,
    ):
        custom_icon = "WORLD"
        if self.layout_type in {"DEFAULT", "COMPACT"}:
            row = layout.row()
            split = row.split(factor=0.1)
            col = split.column()
            col.label(text=f"{index + 1}. ")
            col = split.column()
            col.prop(item, "label", text="", emboss=False)
            hide_icon = "HIDE_OFF" if item.visible else "HIDE_ON"
            row.prop(item, "visible", icon_only=True, icon=hide_icon)
        elif self.layout_type in {"GRID"}:
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)

    def filter_items(self, context, data, propname):
        if data is None:
            return [], []
        items = getattr(data, propname)
        helper_funcs = bpy.types.UI_UL_list
        filtered = []
        # Filtering by name
        if self.filter_name:
            filtered = helper_funcs.filter_items_by_name(
                self.filter_name,
                self.bitflag_filter_item,
                items,
                "label",
                reverse=False,
            )
        if not filtered:
            filtered = [self.bitflag_filter_item] * len(items)
        return filtered, []


class MN_PT_Annotations(bpy.types.Panel):
    """
    Panel for annotations
    """

    bl_idname = "MN_PT_annotations"
    bl_label = "Annotations"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Molecular Nodes"

    @classmethod
    def poll(cls, context):
        """Visible only if entity selected is a trajectory or molecule"""
        scene = context.scene
        active_index = scene.mn.entities_active_index
        if active_index == -1:
            return False
        uuid = scene.mn.entities[active_index].name
        try:
            return scene.MNSession.get(uuid).object.mn.entity_type in (
                EntityType.MD.value,
                EntityType.MD_STREAMING.value,
                EntityType.MOLECULE.value,
                EntityType.DENSITY.value,
            )
        except (LinkedObjectError, AttributeError):
            return False

    def draw(self, context):
        scene = context.scene
        entities_active_index = scene.mn.entities_active_index
        uuid = scene.mn.entities[entities_active_index].name
        entity = scene.MNSession.get(uuid)
        object = entity.object
        annotations_active_index = object.mn.annotations_active_index
        valid_selection = annotations_active_index != -1

        layout = self.layout
        row = layout.row()
        row.prop(object.mn, "annotations_visible", text="Visible")

        row = layout.row()
        MN_UL_AnnotationsList.seqno = 1
        row.template_list(
            "MN_UL_AnnotationsList",
            "annotations_list",
            object,
            "mn_annotations",
            object.mn,
            "annotations_active_index",
            rows=3,
        )
        col = row.column()
        row = col.row()
        op = row.operator("mn.add_annotation", icon="ADD", text="")
        op.uuid = uuid
        row = col.row()
        op = row.operator("mn.remove_annotation", icon="REMOVE", text="")
        if valid_selection:
            op.uuid = uuid
            op.annotation_uuid = object.mn_annotations[annotations_active_index].name
        else:
            row.enabled = False

        if not valid_selection:
            return

        item = object.mn_annotations[annotations_active_index]
        box = layout.box()
        row = box.row()
        row.prop(item, "type")
        row.enabled = False
        entity_annotation_type = f"{entity._mn_entity_type}_{item.type}"
        inputs = getattr(item, entity_annotation_type, None)
        instance = entity.annotations._interfaces.get(inputs.uuid)._instance
        if inputs is not None:
            if not inputs.valid_inputs:
                col = layout.column()
                box = col.box()
                box.label(text="Invalid inputs", icon="ERROR")
                box.alert = True

            for prop_name in inputs.__annotations__.keys():
                if prop_name in ("uuid", "valid_inputs"):
                    continue
                row = box.row()
                if hasattr(instance, f"_{prop_name}"):
                    # indicate use of non blender property in draw
                    row.label(icon="ERROR")
                    row.alert = True
                row.prop(inputs, prop_name)

        # Add all the common annotation params within the 'Options' panel
        header, panel = box.panel("annotation_options", default_closed=True)
        header.label(text="Options")

        if panel:
            text_header, text_panel = panel.panel("text_options", default_closed=True)
            text_header.label(text="Text")
            line_header, line_panel = panel.panel("line_options", default_closed=True)
            line_header.label(text="Lines")
            mesh_header, mesh_panel = panel.panel("mesh_options", default_closed=True)
            mesh_header.label(text="Meshes")

            for prop in item.bl_rna.properties:
                if not prop.is_runtime:
                    continue
                if prop.identifier in ("label", "type", "visible"):
                    continue
                if prop.type == "POINTER" and prop.identifier != "mesh_material":
                    continue
                if text_panel and prop.identifier.startswith("text_"):
                    row = text_panel.row()
                    row.prop(item, prop.identifier)
                    if prop.identifier == "text_falloff":
                        row.enabled = item.text_depth
                elif line_panel and prop.identifier.startswith("line_"):
                    row = line_panel.row()
                    row.prop(item, prop.identifier)
                    if prop.identifier == "line_mode":
                        continue
                    row.enabled = item.line_mode != "mesh"
                elif mesh_panel and prop.identifier.startswith("mesh_"):
                    row = mesh_panel.row()
                    row.prop(item, prop.identifier)

        row = box.row()


class MN_PT_Compositor(bpy.types.Panel):
    """
    Panel for Compositor
    """

    bl_idname = "MN_PT_compositor"
    bl_label = "Options"
    bl_space_type = "NODE_EDITOR"
    bl_region_type = "UI"
    bl_category = "Molecular Nodes"

    @classmethod
    def poll(cls, context):
        return (
            context.area.type == "NODE_EDITOR"
            and context.space_data.tree_type == "CompositorNodeTree"
        )

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        row = box.row()
        row.operator("mn.setup_compositor")


CLASSES = [
    MN_PT_Scene,
    MN_UL_EntitiesList,
    MN_PT_Entities,
    MN_PT_trajectory,
    MN_UL_StylesList,
    MN_PT_Styles,
    MN_UL_AnnotationsList,
    MN_PT_Annotations,
    MN_PT_Compositor,
]



import io
import warnings
from abc import ABCMeta
from pathlib import Path
from typing import Callable
import biotite.structure as struc
import bpy
import databpy
import numpy as np
from biotite import InvalidFileError
from biotite.structure import AtomArray, AtomArrayStack
from ... import blender as bl
from ... import download, utils
from ...nodes import nodes
from ...nodes.geometry import (
    add_style_branch,
)
from ...nodes.nodes import styles_mapping
from ...nodes.styles import (
    StyleBase,
)
from ..base import EntityType, MolecularEntity
from ..utilities import create_object
from . import pdb, pdbx, sdf, selections
from .annotations import MoleculeAnnotationManager
from .reader import ReaderBase


class Molecule(MolecularEntity, metaclass=ABCMeta):
    """
    Primary Molecular Nodes class that coordinates the conversion of structural bioinformatic data
    into raw Blender data.  Most notable the conversion of atoms and bonds into a collection
    of vertices and lines.

    It associates the atomic data (the array) with the created 3D model inside of Blender
    (the object). If multiple conformations are imported, then a `frames` collection
    is also instantiated.

    The `named_attribute()` and `store_named_attribute()` methods access and set attributes on
    `object` that is in the Blender scene.

    Attributes
    ----------
    object : bpy.types.Object
        The Blender object representing the molecule.
    frames : bpy.types.Collection
        The Blender collection which holds the objects making up the frames to animate.
    array: AtomArray | AtomArrayStack:
        The numpy array which stores the atomic coordinates and associated attributes.
    select: MoleculeSelector
        A selector object that provides methods for creating atom selections based on various
        criteria such as atom name, residue type, chain ID, etc. These selections can be used
        with the `add_style` method to apply visual styles to specific parts of the molecule.
    """

    def __init__(
        self,
        array: AtomArray | AtomArrayStack,
        reader: ReaderBase | None = None,
    ):
        """
        Initialize the Molecule object.

        Parameters
        ----------
        array : AtomArray | AtomArrayStack
            The atom array or atom array stack containing the molecular data.
        reader : ReaderBase | None, optional
            The reader used to load the molecular data, by default None.
        """
        self._frames_collection: str | None = None
        self._code: str | None = None
        self._entity_type = EntityType.MOLECULE
        self._reader: ReaderBase | None = reader
        super().__init__()
        self.array = array
        self.select = MoleculeSelector(self)
        self.annotations = MoleculeAnnotationManager(self)

    def create_object(self, name: str = "NewObject"):
        """
        Create a 3D model of the molecule, with one vertex for each atom.
        """
        self.object = create_object(
            array=self.array,
            name=name,
            collection=bl.coll.mn(),
        )
        self.object.mn.entity_type = self._entity_type.value
        if self._reader is not None:
            self._store_object_custom_properties(self.object, self._reader)
        self._setup_frames_collection()
        self._setup_modifiers()

    @classmethod
    def load(
        cls, file_path: str | Path, name: str | None = None, remove_solvent: bool = True
    ) -> "Molecule":
        """
        Load a molecule from a file.

        Parameters
        ----------
        file_path : str or Path
            The path to the file containing molecular data
        name : str or None, optional
            The name to give the molecule object. If None, uses the filename stem
        remove_solvent : bool, optional
            Whether to remove solvent molecules from the structure, default True

        Returns
        -------
        mol : Molecule
            The loaded molecule object with associated data and 3D representation

        Notes
        -----
        Supports various file formats including .cif, .bcif, .pdb, .sdf, and .mol
        """
        reader = cls._read(file_path)
        if not name:
            name = Path(file_path).stem
        mol = cls(reader.array, reader=reader)

        if remove_solvent:
            if isinstance(mol.array, AtomArrayStack):
                mol.array = mol.array[:, ~struc.filter_solvent(mol.array)]
            else:
                mol.array = mol.array[~struc.filter_solvent(mol.array)]

        mol.create_object(name=name)
        mol._reader = reader

        try:
            mol._assemblies = reader._assemblies()
        except InvalidFileError:
            mol._assemblies = ""

        return mol

    @property
    def code(self) -> str | None:
        """
        Get the code for the molecule.
        """
        return self._code

    @staticmethod
    def _read(
        file_path: str | Path | io.BytesIO,
        # remove_solvent: bool = False,
        # del_hydrogen: bool = False,
    ) -> ReaderBase:
        """
        Initially open the file, ready to extract the required data.

        Parameters
        ----------
        file_path : Union[Path, io.BytesIO]
            The file path to the file which stores the atomic coordinates.
        """
        if isinstance(file_path, io.BytesIO):
            reader = pdbx.PDBXReader(file_path)
        else:
            if isinstance(file_path, str):
                file_path = Path(file_path)

            match file_path.suffix:
                case ".cif":
                    reader = pdbx.PDBXReader(file_path)
                case ".bcif":
                    reader = pdbx.PDBXReader(file_path)
                case ".pdb":
                    reader = pdb.PDBReader(file_path)
                case ".sdf":
                    reader = sdf.SDFReader(file_path)
                case ".mol":
                    reader = sdf.SDFReader(file_path)
                case _:
                    raise InvalidFileError("The file format is not supported.")

        return reader

    @classmethod
    def fetch(
        cls,
        code: str,
        format=".bcif",
        centre: str | None = None,
        remove_solvent: bool = True,
        cache: Path | str | None = download.CACHE_DIR,
        database: str = "rcsb",
    ):
        """
        Fetch a molecule from the RCSB database.

        Parameters
        ----------
        code : str
            The PDB ID code of the molecule to fetch.
        format : str, optional
            The file format to download. Default is ".bcif".
        centre : str | None, optional
            Method to use for centering the molecule. Options are "centroid" (geometric center)
            or "mass" (center of mass). If None, no centering is performed. Default is None.
        cache : str, optional
            Path to cache directory. If None, no caching is performed.
        remove_solvent : bool, optional
            Whether to remove solvent from the molecule. Default is True.
        database : str, optional
            The database to fetch from. Default is "rcsb".

        Returns
        -------
        Molecule
            A new Molecule instance created from the downloaded data.
        """
        file_path = download.StructureDownloader(cache=cache).download(
            code=code, format=format, database=database
        )
        mol = cls.load(file_path, name=code, remove_solvent=remove_solvent)
        mol._code = code

        return mol

    def centre_molecule(self, method: str | None = "centroid"):
        """
        Offset positions to centre the atoms and vertices over either the geometric centroid
        or the centre of mass.
        """
        if method is None or method == "":
            return self

        adjustment = self.centroid(weight=method)
        self.position -= adjustment
        self.array.coord -= adjustment
        return self

    def centroid(self, weight: str | np.ndarray | None = None) -> np.ndarray:
        if weight == "centroid" or weight == "":
            return super().centroid()

        return super().centroid(weight)

    @property
    def frames(self) -> bpy.types.Collection | None:
        """
        Get the collection of frames for the molecule.

        Returns
        -------
        bpy.types.Collection
            The collection of frames for the molecule.
        """
        if self._frames_collection is None:
            return None

        return bpy.data.collections[self._frames_collection]

    @frames.setter
    def frames(self, value: bpy.types.Collection):
        """
        Set the collection of frames for the molecule.

        Parameters
        ----------
        value : bpy.types.Collection
            The collection of frames for the molecule.
        """
        if value is None:
            self._frames_collection = None
            return
        if not isinstance(value, bpy.types.Collection):
            raise TypeError("The frames must be a bpy.types.Collection.")

        self._frames_collection = value.name

    @property
    def n_models(self):
        """
        Get the number of models in the molecule.

        Returns
        -------
        int
            The number of models in the molecule.
        """
        if isinstance(self.array, struc.AtomArray):
            return 1
        else:
            return self.array.shape[0]

    def add_style(
        self,
        style: StyleBase | str = "spheres",
        color: str | None = "common",
        selection: "str | MoleculeSelector | None" = None,
        assembly: bool = False,
        material: bpy.types.Material | str | None = None,
        name: str | None = None,
    ):
        """
        Add a visual style to the molecule.

        Parameters
        ----------
        style : bpy.types.GeometryNodeTree | str, optional
            The style to apply to the molecule. Can be a GeometryNodeTree or a string
            identifying a predefined style (e.g., "spheres", "sticks", "ball_stick").
            Default is "spheres".

        color : str | None, optional
            The coloring scheme to apply. Can be "common" (element-based coloring),
            "chain", "residue", or other supported schemes. If None, no coloring
            is applied. Default is "common".

        selection : str | MoleculeSelector | None, optional
            Apply the style only to atoms matching this selection. Can be:
            - A string referring to an existing boolean attribute on the molecule
            - A MoleculeSelector object defining a selection criteria
            - None to apply to all atoms (default)

        assembly : bool, optional
            If True, set up the style to work with biological assemblies.
            Default is False.

        material : bpy.types.Material | str | None, optional
            The material to apply to the styled atoms. Can be a Blender Material object,
            a string with a material name, or None to use default materials. Default is None.

        name: str, optional
            The label for this style

        Returns
        -------
        Molecule
            Returns self for method chaining.

        Raises
        ------
        ValueError
            If an unsupported style string is passed

        Notes
        -----
        If a MoleculeSelector is provided, it will be evaluated and stored as a new
        named attribute on the molecule with an automatically generated name (sel_N).
        """
        if style is None:
            return self

        if isinstance(style, str) and style not in styles_mapping:
            raise ValueError(
                f"Invalid style '{style}'. Supported styles are {[key for key in styles_mapping.keys()]}"
            )

        if isinstance(selection, str) and selection not in self.list_attributes(
            drop_hidden=False
        ):
            warnings.warn(
                f"Named Attribute: '{selection}' does not exist. Style will be added but nothing will be displayed unless that attribute is created.",
                category=UserWarning,
            )

        if isinstance(selection, MoleculeSelector):
            attribute_name = "sel_0"
            i = 0
            while attribute_name in self.list_attributes():
                attribute_name = f"sel_{i}"
                i += 1

            self.store_named_attribute(
                selection.evaluate_on_array(self.array),
                name=attribute_name,
                atype=databpy.AttributeTypes.BOOLEAN,
                domain=databpy.AttributeDomains.POINT,
            )

            selection = attribute_name

        node_style = add_style_branch(
            tree=self.tree,
            style=style,
            color=color,
            selection=selection,
            material=material,
            frames=self.frames,
            name=name,
        )

        # set the active index for UI to the newly added style
        self.object.mn.styles_active_index = self.tree.nodes.find(node_style.name)

        if assembly:
            nodes.assembly_initialise(self.object)
            nodes.assembly_insert(self.object)

        return self

    def _setup_frames_collection(self):
        if self.n_models > 1:
            self.frames = bl.coll.frames(self.name)
            for i, array in enumerate(self.array):
                databpy.create_object(
                    vertices=array.coord * self._world_scale,  # type: ignore
                    name="{}_frame_{}".format(self.name, str(i)),
                    collection=self.frames,
                )

    @staticmethod
    def _store_object_custom_properties(obj, reader: ReaderBase):
        obj["entity_ids"] = reader.entity_ids()
        obj["chain_ids"] = reader.chain_ids()
        obj.mn.biological_assemblies = reader.assemblies(as_json_string=True)

    def assemblies(self, as_array=False):
        """
        Get the biological assemblies of the molecule.

        Parameters
        ----------
        as_array : bool, optional
            Whether to return the assemblies as an array of quaternions. Default is False.

        Returns
        -------
        dict or None
            The biological assemblies of the molecule, as a dictionary of
            transformation matrices, or None if no assemblies are available.
        """
        try:
            if self._reader is None:
                raise InvalidFileError
            assemblies_info = self._reader._assemblies()
        except InvalidFileError:
            return None

        if isinstance(assemblies_info, dict) and as_array:
            return utils.array_quaternions_from_dict(assemblies_info)

        return assemblies_info

    def __repr__(self) -> str:
        """
        Get the string representation of the Molecule object.

        Returns
        -------
        str
            The string representation of the Molecule object.
        """
        return f"<Molecule object: {self.name}>"


class MoleculeSelector:
    """
    A helper to create selections for Molecules and AtomArrays.

    The selection (self.mask) is not computed or returned until the `evaluate_on_array`
    method is called. Until then methods are stored for later evaluation.

    Parameters
    ----------
    mol : Molecule
        The molecule object to select from.

    Attributes
    ----------
    mol : Molecule
        The molecule object to select from.
    mask : ndarray or None
        Boolean array for the selection on the most recently evaluated array.
    pending_selections : list
        List of selection operations to be applied once `evaluate_on_array` is called.
    """

    def __init__(self, mol: Molecule | None = None):
        self.mol = mol
        self.mask: np.ndarray | None = None
        self.pending_selections: list[Callable] = []

    def _update_mask(self, operation):
        """
        Add a selection operation to the pending list.

        Parameters
        ----------
        operation : callable
            Selection operation to add

        Returns
        -------
        self : Selector
            Returns self for method chaining
        """
        self.pending_selections.append(operation)
        return self

    def reset(self):
        """
        Reset all pending selections and the mask

        Returns
        -------
        self : Selector
            Returns self for method chaining
        """
        self.pending_selections = []
        self.mask = None
        return self

    def store_selection(self, name: str) -> None:
        """
        Evaluate and store the current selection as a named attribute on the Molecule

        Parameters
        ----------
        name : str
            The name to store the selection under.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If no selection has been made.
        """
        if self.mol is None:
            raise ValueError("No Molecule associated with this selector")
        if self.mask is None or len(self.pending_selections) > 0:
            self.evaluate_on_array(self.mol.array)

        if self.mask is None:
            raise ValueError("No selection made.")

        self.mol.store_named_attribute(self.mask, name)
        return None

    def evaluate_on_array(self, array: AtomArray | AtomArrayStack) -> np.ndarray:
        """Evaluate this selection on the AtomArray.

        Parameters
        ----------
        array : AtomArray or AtomArrayStack
            The atomic structure to evaluate the selection on.

        Returns
        -------
        ndarray
            Boolean mask array indicating which atoms match the selection criteria.

        Notes
        -----
        All of the selection operations that have been staged for this Selector are
        evaluated and combined with a logical AND, using the AtomArray as input.
        """
        if isinstance(array, AtomArrayStack):
            array = array[0]  # type: ignore

        self.mask = np.ones(array.array_length(), dtype=bool)
        if not self.pending_selections:
            return self.mask

        for operation in self.pending_selections:
            self.mask &= operation(array)
        return self.mask  # type: ignore

    def atom_name(self, atom_name: str | list[str] | tuple[str, ...] | np.ndarray):
        """Select atoms by their name.

        Parameters
        ----------
        atom_name : str or list of str or tuple of str or ndarray
            The atom name(s) to select.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(
            lambda arr: selections.select_atom_names(arr, atom_name)
        )

    def is_amino_acid(self):
        """Select amino acid residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_amino_acids)

    def is_canonical_amino_acid(self):
        """Select canonical amino acid residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_canonical_amino_acids)

    def is_canonical_nucleotide(self):
        """Select canonical nucleotide residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_canonical_nucleotides)

    def is_carbohydrate(self):
        """Select carbohydrate residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_carbohydrates)

    def chain_id(self, chain_id: str | list[str] | tuple[str, ...] | np.ndarray):
        """Select atoms by chain identifier.

        Parameters
        ----------
        chain_id : list of str or tuple of str or ndarray
            The chain identifier(s) to select.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: selections.select_chain_id(arr, chain_id))

    def element(self, element: list[str] | tuple[str, ...] | np.ndarray):
        """Select atoms by element symbol.

        Parameters
        ----------
        element : list of str or tuple of str or ndarray
            The element symbol(s) to select.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: selections.select_element(arr, element))

    def is_hetero(self):
        """Select hetero atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_hetero)

    def is_ligand(self):
        """Select ligand atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_ligand)

    def linear_bond_continuity(self):
        """Select atoms with linear bond continuity.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_linear_bond_continuity)

    def is_monoatomic_ion(self):
        """Select monoatomic ions.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_monoatomic_ions)

    def is_nucleotide(self):
        """Select nucleotide residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_nucleotides)

    def is_peptide_backbone(self):
        """Select peptide backbone atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_peptide_backbone)

    def is_phosphate_backbone(self):
        """Select phosphate backbone atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_phosphate_backbone)

    def is_backbone(self):
        """Select backbone atoms for peptide and nucleotide."""
        return self._update_mask(selections.select_backbone)

    def is_peptide(self):
        return self._update_mask(selections.select_peptide)

    def is_side_chain(self):
        return self._update_mask(selections.select_side_chain)

    def is_polymer(self):
        """Select polymer atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_polymer)

    def res_id(self, num):
        """Select atoms by residue ID.

        Parameters
        ----------
        num : int or list of int
            The residue ID(s) to select.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: selections.select_res_id(arr, num))

    def res_name(self, res_name):
        """Select atoms by residue name.

        Parameters
        ----------
        res_name : str or list of str
            The residue name(s) to select.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: selections.select_res_name(arr, res_name))

    def is_solvent(self):
        """Select solvent atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(selections.select_solvent)

    def not_amino_acids(self):
        """Select non-amino acid residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_amino_acids(arr))

    def not_atom_names(self, atomname):
        """Select atoms not matching the specified atom names.

        Parameters
        ----------
        atomname : str or list of str
            The atom name(s) to exclude.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(
            lambda arr: ~selections.select_atom_names(arr, atomname)
        )

    def not_canonical_amino_acids(self):
        """Select non-canonical amino acid residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(
            lambda arr: ~selections.select_canonical_amino_acids(arr)
        )

    def not_canonical_nucleotides(self):
        """Select non-canonical nucleotide residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(
            lambda arr: ~selections.select_canonical_nucleotides(arr)
        )

    def not_carbohydrates(self):
        """Select non-carbohydrate residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_carbohydrates(arr))

    def not_chain_id(self, chain_id):
        """Select atoms not in the specified chains.

        Parameters
        ----------
        chain_id : str or list of str
            The chain identifier(s) to exclude.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_chain_id(arr, chain_id))

    def not_element(self, element):
        """Select atoms not matching the specified elements.

        Parameters
        ----------
        element : str or list of str
            The element symbol(s) to exclude.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_element(arr, element))

    def not_hetero(self):
        """Select non-hetero atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_hetero(arr))

    def not_monoatomic_ions(self):
        """Select non-monoatomic ion atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_monoatomic_ions(arr))

    def not_peptide(self):
        return self._update_mask(lambda arr: ~selections.select_peptide(arr))

    def not_side_chain(self):
        return self._update_mask(lambda arr: ~selections.select_side_chain(arr))

    def not_nucleotides(self):
        """Select non-nucleotide residues.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_nucleotides(arr))

    def not_peptide_backbone(self):
        """Select non-peptide backbone atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_peptide_backbone(arr))

    def not_phosphate_backbone(self):
        """Select non-phosphate backbone atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_phosphate_backbone(arr))

    def not_polymer(self):
        """Select non-polymer atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_polymer(arr))

    def not_res_id(self, num):
        """Select atoms not matching the specified residue IDs.

        Parameters
        ----------
        num : int or list of int
            The residue ID(s) to exclude.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_res_id(arr, num))

    def not_res_name(self, res_name):
        """Select atoms not matching the specified residue names.

        Parameters
        ----------
        res_name : str or list of str
            The residue name(s) to exclude.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_res_name(arr, res_name))

    def not_solvent(self):
        """Select non-solvent atoms.

        Returns
        -------
        Selector
            Returns self for method chaining.
        """
        return self._update_mask(lambda arr: ~selections.select_solvent(arr))



import bpy
import MDAnalysis as mda
import pytest
import molecularnodes as mn
from .constants import data_dir


def test_session_present():
    assert isinstance(mn.session.get_session(), mn.session.MNSession)


def test_persistent_handlers_added():
    load_handlers = [handler.__name__ for handler in bpy.app.handlers.load_post]
    save_handlers = [handler.__name__ for handler in bpy.app.handlers.save_post]
    assert "_pickle" in save_handlers
    assert "_load" in load_handlers


def test_entity_registered():
    session = mn.session.get_session()
    assert len(session.entities) == 0

    mol = mn.Molecule.fetch("1BNA", cache=data_dir)

    assert mol.uuid in session.entities
    assert isinstance(session.get(mol.uuid), mn.Molecule)
    assert len(session.entities) == 1


@pytest.fixture()
def session():
    return mn.session.get_session()


@pytest.fixture()
def universe():
    topo = data_dir / "md_ppr/box.gro"
    traj = data_dir / "md_ppr/first_5_frames.xtc"
    return mda.Universe(topo, traj)


@pytest.mark.filterwarnings("ignore:.*Empty string to select atoms.*:UserWarning")
def test_add_trajectory(session, universe):
    # add Universe as trajectory
    t1 = session.add_trajectory(universe, name="u1")
    assert "u1" in bpy.data.objects
    assert t1._mn_entity_type == mn.entities.base.EntityType.MD.value
    # add AtomGroup as trajectory
    ag = universe.select_atoms("name CA")
    session.add_trajectory(ag, name="ag1")
    assert "ag1" in bpy.data.objects


def test_remove_trajectory(session, universe):
    t1 = session.add_trajectory(universe, name="u1")
    assert "u1" in bpy.data.objects
    # remove by trajectory instance
    session.remove_trajectory(t1)
    assert "u1" not in bpy.data.objects
    session.add_trajectory(universe, name="u2")
    assert "u2" in bpy.data.objects
    # remove by trajectory name
    session.remove_trajectory("u2")
    t3 = session.add_trajectory(universe, name="u3")
    assert "u3" in bpy.data.objects
    # remove trajectory from UI using operator
    bpy.ops.mn.session_remove_item("EXEC_DEFAULT", uuid=t3.uuid)
    assert "u3" not in bpy.data.objects


def test_get_trajectory(session, universe):
    t1 = session.add_trajectory(universe, name="u1")
    t2 = session.get_trajectory("u1")
    assert t1 == t2


def test_entity_blender_properties(session, universe):
    props = bpy.context.scene.mn
    assert len(props.entities) == 0
    t1 = session.add_trajectory(universe, name="u1")
    # test property addition
    assert len(props.entities) == 1
    entity = props.entities[0]
    # test property indexing
    assert entity.name == t1.uuid
    # verify entity type
    assert entity.type == "md"
    # test visibility changes
    # check initial visibility
    assert entity.visible
    assert bpy.data.objects["u1"].visible_get()
    entity.visible = False
    assert not bpy.data.objects["u1"].visible_get()
    # test property removal
    session.remove_trajectory("u1")
    assert len(props.entities) == 0



import os
import sys
from os.path import dirname, join, realpath
import bpy
import pytest
import molecularnodes as mn
from .utils import NumpySnapshotExtension

mn.ui.addon._test_register()


DATA_DIR = join(dirname(realpath(__file__)), "data")
IS_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"
IS_SELF_HOSTED = os.getenv("environment") == "self-hosted"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before each tests

    bpy.ops.wm.read_homefile(app_template="")
    mn.session.get_session().clear()
    for tree in bpy.data.node_groups:
        bpy.data.node_groups.remove(tree)

    print(f"{mn.session.get_session().entities=}")
    print(f"{list(bpy.data.objects)=}")

    yield
    print("Post Test setup:")
    print(f"{bpy.app.handlers.frame_change_pre=}")
    print(f"{mn.session.get_session().entities.keys()=}")
    print(f"{list(bpy.data.objects)=}")
    print(f"{list(o.uuid for o in bpy.data.objects)=}")
    bpy.ops.wm.read_homefile(app_template="")
    mn.session.get_session().clear()
    # Code that will run after your test, for example:
    # files_after = # ... do something to check the existing files
    # assert files_before == files_after


def pytest_sessionstart(session):
    """
    Insert ``MolecularNodes`` into ``PYTHONPATH`` to make it importable.
    """
    project_dir = dirname(dirname(realpath(__file__)))
    sys.path.insert(0, join(project_dir))


@pytest.fixture
def snapshot_custom(snapshot):
    return snapshot.use_extension(NumpySnapshotExtension)



import gzip
import io
from pathlib import Path
import requests
from biotite.database import afdb

CACHE_DIR = Path(Path.home(), "MolecularNodesCache").expanduser()


class FileDownloadPDBError(Exception):
    """
    Exception raised for errors in the file download process.

    Attributes
    ----------
    message : str
        Explanation of the error
    """

    def __init__(
        self,
        message="There was an error downloading the file from the Protein Data Bank. PDB or format for PDB code may not be available.",
    ):
        self.message = message
        super().__init__(self.message)


class StructureDownloader:
    """
    A class for downloading molecular structure files from various databases.

    Parameters
    ----------
    cache : str, Path, or None, optional
        Directory path for caching downloaded files. If None, files are not cached.
        Defaults to CACHE_DIR.
    """

    def __init__(self, cache: str | Path | None = CACHE_DIR):
        if cache:
            cache_path = Path(cache).absolute()
            cache_path.mkdir(parents=True, exist_ok=True)
            self.cache = cache_path
        else:
            self.cache = None

    def download(
        self,
        code: str,
        format: str = "cif",
        database: str = "rcsb",
    ) -> Path | io.BytesIO | io.StringIO:
        """Download a structure from the specified protein data bank in the given format.

        Parameters
        ----------
        code : str
            The code of the file to fetch. Supports both traditional 4-character codes
            and new format codes with 'pdb_' prefix (e.g., 'pdb_00009bdt').
        format : str, optional
            The format of the file. Must be one of ['cif', 'pdb', 'bcif'].
            Defaults to "cif".
        database : str, optional
            The database to fetch the file from. Must be one of ['rcsb', 'pdb', 'wwpdb', 'alphafold'].
            Defaults to 'rcsb'.

        Returns
        -------
        Path or io.BytesIO or io.StringIO
            If cache is enabled, returns the path to the cached file as a Path.
            If cache is disabled, returns either:
            - io.BytesIO for binary formats (bcif)
            - io.StringIO for text formats (cif, pdb)

        Raises
        ------
        ValueError
            If the specified format is not supported, or if new format PDB codes
            are used with PDB format.
        FileDownloadPDBError
            If there is an error downloading the file from the database.

        Examples
        --------
        >>> downloader = StructureDownloader()
        >>> path = downloader.download("1abc", format="cif")
        >>> path = downloader.download("pdb_00009bdt", format="bcif")
        """
        code = code.strip()
        format = format.strip(".")
        supported_formats = ["cif", "pdb", "bcif"]
        if format not in supported_formats:
            raise ValueError(f"File format '{format}' not in: {supported_formats=}")

        # Check if the code has the new format prefix and is requesting PDB format
        if code.startswith("pdb_") and format == "pdb":
            raise ValueError(
                "New format PDB codes (starting with 'pdb_') are not compatible with .pdb format. Please use 'cif' or 'bcif' format instead."
            )

        # Use biotite's afdb.fetch for AlphaFold database
        if database == "alphafold":
            try:
                result = afdb.fetch(
                    code, format=format, target_path=self.cache, overwrite=False
                )
                # biotite returns file path as string if cache is used, or StringIO/BytesIO if not
                if isinstance(result, str):
                    return Path(result)
                return result
            except Exception as e:
                raise FileDownloadPDBError(
                    f"Error downloading from AlphaFold database: {str(e)}"
                )

        _is_binary = format in ["bcif"]
        filename = f"{code}.{format}"

        if self.cache:
            file = self.cache / filename
            if file.exists():
                return file
        else:
            file = None

        try:
            r = requests.get(self._url(code, format, database))
            r.raise_for_status()
        except requests.HTTPError as e:
            raise FileDownloadPDBError(str(e))

        if _is_binary:
            content = r.content
            # Check if the content is gzipped
            if content[:2] == b"\x1f\x8b":  # gzip magic number
                content = gzip.decompress(content)
        else:
            content = r.text

        if file:
            mode = "wb+" if _is_binary else "w+"
            with open(file, mode) as f:
                f.write(content)
            return Path(file)
        else:
            if _is_binary:
                if not isinstance(content, bytes):
                    raise ValueError(
                        "Binary content is not bytes, please check your format."
                    )
                file = io.BytesIO(content)
            else:
                if not isinstance(content, str):
                    raise ValueError(
                        "Text content is not str, please check your format."
                    )
                file = io.StringIO(content)

        return file

    def _url(self, code: str, format: str, database: str = "rcsb") -> str:
        """Get the URL for downloading the given file from a particular database.

        Parameters
        ----------
        code : str
            The structure code to download.
        format : str
            The file format ('cif', 'pdb', or 'bcif').
        database : str, optional
            The database name. Defaults to 'rcsb'.

        Returns
        -------
        str
            The complete URL for downloading the file.

        Raises
        ------
        ValueError
            If the database is not currently supported.
        """

        if database in ["rcsb", "pdb", "wwpdb"]:
            if format == "bcif":
                return f"https://models.rcsb.org/{code}.bcif"
            else:
                return f"https://files.rcsb.org/download/{code}.{format}"
        else:
            raise ValueError(f"Database {database} not currently supported.")



import subprocess
import sys


def install_dependency_group(pyproject_path, group_name):
    """
    Install packages from a specific dependency group in pyproject.toml.

    Args:
        pyproject_path: Path to pyproject.toml file
        group_name: Name of the dependency group to install
    """
    python_exe = sys.executable
    subprocess.check_call([python_exe, "-m", "pip", "install", "uv"])
    subprocess.check_call([python_exe, "-m", "pip", "install", "-e", "."])
    subprocess.check_call(
        [
            python_exe,
            "-m",
            "uv",
            "pip",
            "install",
            "-r",
            pyproject_path,
            "--extra",
            group_name,
        ]
    )


# Configuration
PYPROJECT_PATH = "pyproject.toml"  # Change this to your file path
GROUP_NAME = "test"  # Change this to your group name (e.g., "dev", "test", "docs")

if __name__ == "__main__":
    install_dependency_group(PYPROJECT_PATH, GROUP_NAME)



import os
import pickle as pk
from typing import Dict, Union
import bpy
import MDAnalysis as mda
from bpy.app.handlers import persistent  # type: ignore
from bpy.props import StringProperty  # type: ignore
from bpy.types import Context  # type: ignore
from databpy.object import LinkedObjectError, get_from_uuid
from MDAnalysis.core.groups import AtomGroup
from .entities import Molecule
from .entities.base import EntityType, MolecularEntity
from .entities.ensemble.base import Ensemble
from .entities.trajectory.base import Trajectory
from .nodes.nodes import styles_mapping


def trim(dictionary: dict):
    dic = dictionary.copy()
    for key in list(dic.keys()):
        if dic[key].object is None:
            dic.pop(key)
    return dic


def make_paths_relative(trajectories: Dict[str, Trajectory]) -> None:
    for key, traj in trajectories.items():
        # save linked universe frame
        uframe = traj.uframe
        traj.universe.load_new(make_path_relative(traj.universe.trajectory.filename))
        # restore linked universe frame
        traj.uframe = uframe
        traj._save_filepaths_on_object()


def trim_root_folder(filename):
    "Remove one of the prefix folders from a filepath"
    return os.sep.join(filename.split(os.sep)[1:])


def make_path_relative(filepath):
    "Take a path and make it relative, in an actually usable way"
    try:
        filepath = os.path.relpath(filepath)
    except ValueError:
        return filepath

    # count the number of "../../../" there are to remove
    n_to_remove = int(filepath.count("..") - 2)
    # get the filepath without the huge number of "../../../../" at the start
    sans_relative = filepath.split("..")[-1]

    if n_to_remove < 1:
        return filepath

    for i in range(n_to_remove):
        sans_relative = trim_root_folder(sans_relative)

    return f"./{sans_relative}"


class MNSession:
    def __init__(self) -> None:
        self.entities: Dict[str, Union[Molecule, Trajectory, Ensemble]] = {}

    @property
    def molecules(self) -> Dict[str, Molecule]:
        return {k: v for k, v in self.entities.items() if isinstance(v, Molecule)}

    @property
    def trajectories(self) -> Dict[str, Trajectory]:
        # return a filtered dictionary of only the trajectories using isinstance(item, Trajectory)
        return {k: v for k, v in self.entities.items() if isinstance(v, Trajectory)}

    @property
    def ensembles(self) -> Dict[str, Ensemble]:
        # return a filtered dictionary of only the ensembles using isinstance(item, Ensemble)
        return {k: v for k, v in self.entities.items() if isinstance(v, Ensemble)}

    def register_entity(self, item: MolecularEntity) -> None:
        """Add entity to the dictionary"""
        self.entities[item.uuid] = item
        # add entity to blender properties if it doesn't exist
        props = bpy.context.scene.mn
        entities = props.entities  # entities collection
        if entities.find(item.uuid) == -1:
            entity = entities.add()  # add new entity to collection
            entity.name = item.uuid  # immutable name that allows find to get index
            # EntityType is not available here in most cases as it is set after __init__
            # In other cases, it is reset later like in case of MD_OXDNA
            # So, use a simple check based on supported entity types here
            if isinstance(item, Molecule):
                entity.type = EntityType.MOLECULE.value
            elif isinstance(item, Trajectory):
                entity.type = EntityType.MD.value
            elif isinstance(item, Ensemble):
                entity.type = EntityType.ENSEMBLE.value
            props.entities_active_index = len(entities) - 1

    def remove_entity(self, uuid: str) -> None:
        """Remove entity from the dictionary"""
        del self.entities[uuid]
        # remove entity from blender properties
        props = bpy.context.scene.mn
        entities = props.entities
        index = entities.find(uuid)
        if index != -1:
            entities.remove(index)  # remove entity from collection
            props.entities_active_index = len(entities) - 1

    def match(self, obj: bpy.types.Object) -> Union[Molecule, Trajectory, Ensemble]:
        return self.get(obj.uuid)

    def get_object(self, uuid: str) -> bpy.types.Object | None:
        """
        Try and get an object from Blender's object database that matches the uuid given.

        If nothing is be found to match, return None.
        """
        return get_from_uuid(uuid)

    def get(self, uuid: str) -> Union[Molecule, Trajectory, Ensemble] | None:
        return self.entities.get(uuid)

    def prune(self) -> None:
        """
        Remove any entities that no longer exist in Blender
        """
        # remove any entities that don't have linked objects
        for uuid in list(self.entities):
            try:
                _ = self.entities[uuid].name
            except LinkedObjectError:
                self.remove_entity(uuid)
        # remove any properties that don't exist in session
        props = bpy.context.scene.mn
        entities = props.entities
        remove_indices = [
            i for i, entity in enumerate(entities) if entity.name not in self.entities
        ]
        if remove_indices:
            # remove indices in the reverse sorted order for correctness
            reversed_indices = sorted(remove_indices, reverse=True)
            for i in reversed_indices:
                if i < len(entities):
                    entities.remove(i)
            props.entities_active_index = len(entities) - 1

    @property
    def n_items(self) -> int:
        "The number of items being tracked by this session."
        return len(self.entities)

    def __repr__(self) -> str:
        return f"MNSession with {len(self.molecules)} molecules, {len(self.trajectories)} trajectories and {len(self.ensembles)} ensembles."

    def pickle(self, filepath) -> None:
        pickle_path = self.stashpath(filepath)

        make_paths_relative(self.trajectories)
        self.entities = trim(self.entities)

        # don't save anything if there is nothing to save
        if self.n_items == 0:
            return None

        with open(pickle_path, "wb") as f:
            pk.dump(self, f)

        print(f"Saved session to: {pickle_path}")

    def load(self, filepath) -> None:
        pickle_path = self.stashpath(filepath)
        if not os.path.exists(pickle_path):
            raise FileNotFoundError(f"MNSession file `{pickle_path}` not found")
        with open(pickle_path, "rb") as f:
            session = pk.load(f)

        # TODO: clear up in later versions
        # this handles reloading sessions which don't have the `entities` attribute
        if hasattr(session, "entities"):
            for item in session.entities.values():
                self.register_entity(item)
        else:
            for mol in session.molecules.values():
                self.register_entity(mol)

            for traj in session.trajectories.values():
                self.register_entity(traj)

            for ens in session.ensembles.values():
                self.register_entity(ens)

        print(f"Loaded a MNSession from: {pickle_path}")

    def stashpath(self, filepath) -> str:
        return f"{filepath}.MNSession"

    def clear(self) -> None:
        """Remove references to all molecules, trajectories and ensembles."""
        self.entities = {}

    def remove(self, uuid: str) -> None:
        """Remove an entity by uuid"""
        if uuid not in self.entities:
            raise ValueError(f"No entity with UUID '{uuid}'")
        entity = self.entities[uuid]
        bpy.data.objects.remove(entity.object, do_unlink=True)
        if hasattr(entity, "annotations"):
            if entity.annotations.bob:
                bpy.data.objects.remove(entity.annotations.bob.object, do_unlink=True)
        self.remove_entity(uuid)

    def add_trajectory(
        self,
        universe: mda.Universe | AtomGroup,
        style: str | None = "vdw",
        name: str = "NewUniverseObject",
    ) -> Trajectory:
        """
        Add a new trajectory

        Parameters
        ----------
        universe: mda.Universe | AtomGroup, required
            MDAnalysis Universe or AtomGroup instance

        style: str | None, optional
            The style to apply to the Universe or AtomGroup.

        name: str, optional
            Name of the trajectory object in Blender

        Returns
        -------
        Trajectory
            The newly added Trajectory instance

        """
        if style is not None and style not in styles_mapping:
            raise ValueError(
                f"Invalid style '{style}'. Supported styles are {[key for key in styles_mapping.keys()]}"
            )
        selection = None
        if isinstance(universe, AtomGroup):
            traj = Trajectory(universe.universe, name=name)  # AtomGroup universe
            selection = universe  # AtomGroup
        else:
            traj = Trajectory(universe, name=name)  # Universe
        traj.add_style(style=style, selection=selection)
        return traj

    def get_trajectory(
        self,
        name: str,
    ) -> Trajectory:
        """
        Get trajectory instance by name

        Parameters
        ----------
        name: str, required
            Name of the trajectory object

        Returns
        -------
        Trajectory
            A Trajectory instance

        Raises
        ------
        ValueError if trajectory is not found

        """
        for v in self.entities.values():
            if isinstance(v, Trajectory) and v.object.name == name:
                return v
        raise ValueError(f"No trajectory named '{name}'")

    def remove_trajectory(
        self,
        trajectory: Trajectory | str,
    ) -> None:
        """
        Remove an existing trajectory

        Parameters
        ----------
        trajectory: Trajectory | str, required
            A Trajectory instance or name of the trajectory

        Returns
        -------
        None

        Raises
        ------
        ValueError if trajectory name is not found

        """
        instance = trajectory
        if isinstance(trajectory, str):
            instance = self.get_trajectory(trajectory)
        self.remove(instance.uuid)


def get_session(context: Context | None = None) -> MNSession:
    if not context:
        context = bpy.context
    return context.scene.MNSession


def get_entity(context: Context | None = None) -> Molecule | Trajectory | Ensemble:
    session = get_session(context)
    return session.match(context.active_object)


@persistent
def _pickle(filepath) -> None:
    get_session().pickle(filepath)


@persistent
def _load(filepath: str, printing: str = "quiet") -> None:
    """
    Load a session from the specified file path.

    This function attempts to load a session from the given file path using the
    `get_session().load(filepath)` method. If the file path is empty, the function
    returns immediately without attempting to load anything. If the file is not found,
    it handles the `FileNotFoundError` exception and optionally prints a message
    based on the `printing` parameter.

    Args:
        filepath (str): The path to the file from which to load the session. If this
            is an empty string, the function will return without doing anything.
        printing (str, optional): Controls the verbosity of the function. If set to
            "verbose", a message will be printed when the file is not found. Defaults
            to "quiet".

    Returns:
        None: This function does not return any value.

    Raises:
        FileNotFoundError: If the file specified by `filepath` does not exist and
            `printing` is set to "verbose", a message will be printed.
    """
    # the file hasn't been saved or we are opening a fresh file, so don't
    # attempt to load anything
    if filepath == "":
        return None
    try:
        get_session().load(filepath)
    except FileNotFoundError:
        if printing == "verbose":
            print("No MNSession found to load for this .blend file.")
        else:
            pass


class MN_OT_Session_Remove_Item(bpy.types.Operator):
    bl_idname = "mn.session_remove_item"
    bl_label = "Remove"
    bl_description = "Remove this item from the internal Molecular Nodes session"
    bl_options = {"REGISTER", "UNDO"}

    uuid: StringProperty()  # type: ignore

    def invoke(self, context: Context, event):
        session = get_session()

        return context.window_manager.invoke_confirm(
            self,
            event=event,
            title="Permanently delete item?",
            message=f"Any links to objects that rely upon this item will be lost.  {session.get(self.uuid)}",
        )

    def execute(self, context: Context):
        get_session().remove(self.uuid)

        return {"FINISHED"}


class MN_OT_Session_Create_Object(bpy.types.Operator):
    bl_idname = "mn.session_create_object"
    bl_label = "Create Object"
    bl_description = "Create a new object linked to this item"
    bl_options = {"REGISTER", "UNDO"}

    uuid: StringProperty()  # type: ignore

    def execute(self, context: Context):
        item = get_session().get(self.uuid)
        if item is None:
            self.report({"ERROR"}, f"No item with UUID '{self.uuid}'")
            return {"CANCELLED"}
        item.create_object()
        return {"FINISHED"}


class MN_OT_Session_Prune(bpy.types.Operator):
    bl_idname = "mn.session_prune"
    bl_label = "Session Prune"
    bl_description = "Prune session entities by removing ones that no longer exist"

    def execute(self, context: Context):
        get_session().prune()
        return {"FINISHED"}


CLASSES = [MN_OT_Session_Remove_Item, MN_OT_Session_Create_Object, MN_OT_Session_Prune]



from contextlib import contextmanager
from pathlib import Path
import bmesh
import bpy


def path_resolve(path: str | Path) -> Path:
    if isinstance(path, str):
        return Path(bpy.path.abspath(path))
    elif isinstance(path, Path):
        return Path(bpy.path.abspath(str(path)))
    else:
        raise ValueError(f"Unable to resolve path: {path}")


def set_obj_active(
    obj: bpy.types.Object, context: bpy.types.Context | None = None
) -> None:
    if not context:
        context = bpy.context

    context.view_layer.objects.active = obj  # type: ignore


def set_object_visibility(object: bpy.types.Object, visible: bool) -> None:
    """Set visibility of Blender object"""
    if object.name not in bpy.context.view_layer.objects:
        return
    try:
        object.hide_set(not visible)
        # obj.hide_viewport = hide
        # icon incorrect, but causes blender bug for volumes with above
        object.hide_render = not visible
    except RuntimeError:
        # Keyframing object visibility can at times lead to:
        # RuntimeError: Object can't be hidden because it is not in View Layer 'ViewLayer'!
        pass


def viewport_tag_redraw() -> None:
    """Tag all the viewport areas for a redraw"""
    for window in bpy.context.window_manager.windows:
        for area in window.screen.areas:
            if area.type == "VIEW_3D":
                area.tag_redraw()


def get_viewport_region_from_context(context) -> tuple:
    """Get the 3D viewport region and region data from context"""
    region = context.region
    rv3d = None
    if context.space_data is None:
        return (region, rv3d)
    if not context.space_data.region_quadviews:
        rv3d = context.space_data.region_3d
    else:
        # handle quadview case
        if context.area.type != "VIEW_3D" or context.space_data.type != "VIEW_3D":
            return (region, rv3d)
        i = -1
        for region in context.area.regions:
            if region.type == "WINDOW":
                i += 1
                if context.region == region:
                    break
        else:
            return (region, rv3d)
        rv3d = context.space_data.region_quadviews[i]
    return (region, rv3d)


def get_bounding_box(obj: bpy.types.Object) -> list[tuple]:
    return [co[:] for co in obj.bound_box[:]]


def look_at_object(obj: bpy.types.Object) -> None:
    """Set camera to look at an object"""
    # save current selected objects list
    prev_sel = bpy.context.selected_objects
    # de-select all objects in the scene
    for o in bpy.data.objects:
        o.select_set(False)
    # select this object and set camera view
    obj.select_set(True)
    bpy.ops.view3d.camera_to_view_selected()
    obj.select_set(False)
    # restore previous object selections
    for o in prev_sel:
        o.select_set(True)


def look_at_bbox(bbox: list[tuple]):
    """Set camera to look at a bounding box"""
    # create a temporary mesh from bounding box vertices
    bb_mesh = bpy.data.meshes.new("bb")
    bb_mesh.from_pydata(bbox, [], [])
    bb_mesh.update()
    # create a temporary object from mesh data
    bb_obj = bpy.data.objects.new("temp_bb_object", bb_mesh)
    # link temporary object to scene
    bpy.context.collection.objects.link(bb_obj)
    # frame the temporary object
    look_at_object(bb_obj)
    # remove temporary object from scene
    bpy.data.objects.remove(bb_obj, do_unlink=True)


@contextmanager
def new_bmesh() -> bmesh.types.BMesh:  # type: ignore
    # create a new bmesh
    bm = bmesh.new()
    try:
        # return the new bmesh
        yield bm
    finally:
        # free bmesh
        bm.free()



import itertools
from io import BytesIO
from pathlib import Path
import biotite.structure as struc
import biotite.structure.io.pdbx as pdbx
import numpy as np
from biotite import InvalidFileError
from .reader import ReaderBase


class PDBXReader(ReaderBase):
    def __init__(self, file_path: str | Path | BytesIO):
        self._extra_annotations = {
            "sec_struct": self._get_secondary_structure,
            "entity_id": self._get_entity_id,
        }
        self._extra_fields = ["b_factor", "occupancy", "atom_id"]
        super().__init__(file_path)

    def read(self, file_path):
        match file_path.suffix:
            case ".cif":
                return pdbx.CIFFile.read(file_path)
            case ".bcif":
                return pdbx.BinaryCIFFile.read(file_path)
            case _:
                raise NotImplementedError(
                    f"File type {file_path.suffix} not supported."
                )

    def get_structure(
        self, model: int | None = None, include_bonds: bool = True
    ) -> struc.AtomArray | struc.AtomArrayStack:
        try:
            array = pdbx.get_structure(
                self.file, model=model, extra_fields=self._extra_fields
            )
        except InvalidFileError:
            array = pdbx.get_component(self.file)

        if include_bonds and not array.bonds:
            array.bonds = struc.bonds.connect_via_residue_names(
                array, inter_residue=True
            )

        return array  # type: ignore

    def _assemblies(self):
        return CIFAssemblyParser(self.file).get_assemblies()

    def entity_ids(self):
        try:
            return (
                self.file.block.get("entity")
                .get("pdbx_description")
                .as_array()
                .tolist()
            )
        except AttributeError:
            return None

    @staticmethod
    def _extract_matrices(category):
        matrix_columns = [
            "matrix[1][1]",
            "matrix[1][2]",
            "matrix[1][3]",
            "vector[1]",
            "matrix[2][1]",
            "matrix[2][2]",
            "matrix[2][3]",
            "vector[2]",
            "matrix[3][1]",
            "matrix[3][2]",
            "matrix[3][3]",
            "vector[3]",
        ]

        columns = [category[name].as_array().astype(float) for name in matrix_columns]
        matrices = np.empty((len(columns[0]), 4, 4), float)

        col_mask = np.tile((0, 1, 2, 3), 3)
        row_mask = np.repeat((0, 1, 2), 4)
        for column, coli, rowi in zip(columns, col_mask, row_mask):
            matrices[:, rowi, coli] = column

        return matrices

    @staticmethod
    def _get_entity_id(array, file):
        chain_ids = file.block["entity_poly"]["pdbx_strand_id"].as_array(str)

        # the chain_ids are an array of individual items np.array(['A,B', 'C', 'D,E,F'])
        # which need to be categorised as [1, 1, 2, 3, 3, 3] for their belonging to individual
        # entities

        chains = []
        idx = []
        for i, chain_str in enumerate(chain_ids):
            for chain in chain_str.split(","):
                chains.append(chain)
                idx.append(i)

        # this is how we map the chain_ids and res_names of our entities to their integer
        # representations
        entity_lookup = dict(zip(chains, idx))

        # for the hetero atoms, we need to add a new entity_id into the lookup so that
        # they can be assigned an entity ID
        unique_res_het = np.unique(array.res_name[array.hetero])
        for het in unique_res_het:
            if het not in entity_lookup:
                entity_lookup[het] = max(entity_lookup.values()) + 1

        entity_id_int = np.zeros(len(array.res_name), int)
        for i, res_name in enumerate(array.res_name):
            entity_id_int[i] = entity_lookup.get(
                res_name, entity_lookup[array.chain_id[i]]
            )

        return entity_id_int

    @staticmethod
    def _get_secondary_structure(array, file):
        """
        Get secondary structure information for the array from the file.

        Parameters
        ----------
        array : numpy array
            The array for which secondary structure information is to be retrieved.
        file : object
            The file object containing the secondary structure information.

        Returns
        -------
        numpy array
            A numpy array of secondary structure information, where each element is either 0, 1, 2, or 3.
            - 0: Not a peptide
            - 1: Alpha helix
            - 2: Beta sheet
            - 3: Loop

        Raises
        ------
        KeyError
            If the 'struct_conf' category is not found in the file.
        """

        # get the annotations for the struc_conf category. Provides start and end
        # residues for the annotations. For most files this will only contain the
        # alpha helices, but will sometimes contain also other secondary structure
        # information such as in AlphaFold predictions

        conf = file.block.get("struct_conf")
        if conf:
            starts = conf["beg_auth_seq_id"].as_array().astype(int)
            ends = conf["end_auth_seq_id"].as_array().astype(int)
            chains = conf["end_auth_asym_id"].as_array().astype(str)
            id_label = conf["id"].as_array().astype(str)
        else:
            starts = np.empty(0, dtype=int)
            ends = np.empty(0, dtype=int)
            chains = np.empty(0, dtype=str)
            id_label = np.empty(0, dtype=int)

        # most files will have a separate category for the beta sheets
        # this can just be appended to the other start / end / id and be processed
        # as normalquit
        sheet = file.block.get("struct_sheet_range")
        if sheet:
            starts = np.append(starts, sheet["beg_auth_seq_id"].as_array().astype(int))
            ends = np.append(ends, sheet["end_auth_seq_id"].as_array().astype(int))
            chains = np.append(chains, sheet["end_auth_asym_id"].as_array().astype(str))
            id_label = np.append(id_label, np.repeat("STRN", len(sheet["id"])))

        if not conf and not sheet:
            raise KeyError

        # convert the string labels to integer representations of the SS
        # AH: 1, BS: 2, LOOP: 3

        id_int = np.array([_ss_label_to_int(label) for label in id_label], int)

        # create a lookup dictionary that enables lookup of secondary structure
        # based on the chain_id and res_id values

        lookup = dict()
        for chain in np.unique(chains):
            arrays = []
            mask = chain == chains
            start_sub = starts[mask]
            end_sub = ends[mask]
            id_sub = id_int[mask]

            for start, end, id in zip(start_sub, end_sub, id_sub):
                idx = np.arange(start, end + 1, dtype=int)
                arr = np.zeros((len(idx), 2), dtype=int)
                arr[:, 0] = idx
                arr[:, 1] = 3
                arr[:, 1] = id
                arrays.append(arr)

            lookup[chain] = dict(np.vstack(arrays).tolist())

        # use the lookup dictionary to get the SS annotation based on the chain_id and res_id
        secondary_structure = np.zeros(len(array.chain_id), int)
        for i, (chain, res) in enumerate(zip(array.chain_id, array.res_id)):
            try:
                secondary_structure[i] = lookup[chain].get(res, 3)
            except KeyError:
                secondary_structure[i] = 0

        # assign SS to 0 where not peptide
        secondary_structure[~struc.filter_amino_acids(array)] = 0
        return secondary_structure


def _parse_opers(oper):
    # we want the example '1,3,(5-8)' to expand to (1, 3, 5, 6, 7, 8).
    op_ids = list()

    for group in oper.strip(")").split("("):
        if "," in group:
            for i in group.split(","):
                op_ids.append()

    for group in oper.split(","):
        if "-" not in group:
            op_ids.append(str(group))
            continue

        start, stop = [int(x) for x in group.strip("()").split("-")]
        for i in range(start, stop + 1):
            op_ids.append(str(i))

    return op_ids


def _ss_label_to_int(label):
    if "HELX" in label:
        return 1
    elif "STRN" in label:
        return 2
    else:
        return 3


class CIFAssemblyParser:
    # Implementation adapted from ``biotite.structure.io.pdbx.convert``

    def __init__(self, file_cif):
        self._file = file_cif

    def list_assemblies(self):
        return list(pdbx.list_assemblies(self._file).keys())

    def get_transformations(self, assembly_id):
        assembly_gen_category = self._file.block["pdbx_struct_assembly_gen"]

        struct_oper_category = self._file.block["pdbx_struct_oper_list"]

        if assembly_id not in assembly_gen_category["assembly_id"].as_array(str):
            raise KeyError(f"File has no Assembly ID '{assembly_id}'")

        # Extract all possible transformations indexed by operation ID
        # transformation_dict = _get_transformations(struct_oper_category)
        transformation_dict = _extract_matrices(struct_oper_category)

        # Get necessary transformations and the affected chain IDs
        # NOTE: The chains given here refer to the `label_asym_id` field
        # of the `atom_site` category
        # However, by default `PDBxFile` uses the `auth_asym_id` as
        # chain ID
        matrices = []
        pdb_model_num = -1
        for id, op_expr, asym_id_expr in zip(
            assembly_gen_category["assembly_id"].as_array(str),
            assembly_gen_category["oper_expression"].as_array(str),
            assembly_gen_category["asym_id_list"].as_array(str),
        ):
            pdb_model_num += 1
            # Find the operation expressions for given assembly ID
            # We already asserted that the ID is actually present
            if id != assembly_id:
                continue

            operations = _parse_operation_expression(op_expr)

            affected_chain_ids = asym_id_expr.split(",")

            for i, operation in enumerate(operations):
                # for op_step in operation:
                matrices.append(
                    {
                        "chain_ids": affected_chain_ids,
                        "matrix": transformation_dict[operation[0]].tolist(),
                        "pdb_model_num": pdb_model_num,
                    }
                )

        return matrices

    def get_assemblies(self):
        assembly_dict = {}
        for assembly_id in self.list_assemblies():
            assembly_dict[assembly_id] = self.get_transformations(assembly_id)

        return assembly_dict


def _extract_matrices(category, scale=True):
    matrix_columns = [
        "matrix[1][1]",
        "matrix[1][2]",
        "matrix[1][3]",
        "vector[1]",
        "matrix[2][1]",
        "matrix[2][2]",
        "matrix[2][3]",
        "vector[2]",
        "matrix[3][1]",
        "matrix[3][2]",
        "matrix[3][3]",
        "vector[3]",
    ]

    columns = [category[name].as_array().astype(float) for name in matrix_columns]
    n = 4 if scale else 3
    matrices = np.empty((len(columns[0]), n, 4), float)

    col_mask = np.tile((0, 1, 2, 3), 3)
    row_mask = np.repeat((0, 1, 2), 4)
    for column, coli, rowi in zip(columns, col_mask, row_mask):
        matrices[:, rowi, coli] = column

    return dict(zip(category["id"].as_array(str), matrices))


def _chain_transformations(rotations, translations):
    """
    Get a total rotation/translation transformation by combining
    multiple rotation/translation transformations.
    This is done by intermediately combining rotation matrices and
    translation vectors into 4x4 matrices in the form

    |r11 r12 r13 t1|
    |r21 r22 r23 t2|
    |r31 r32 r33 t3|
    |0   0   0   1 |.
    """
    total_matrix = np.identity(4)
    for rotation, translation in zip(rotations, translations):
        matrix = np.zeros((4, 4))
        matrix[:3, :3] = rotation
        matrix[:3, 3] = translation
        matrix[3, 3] = 1
        total_matrix = matrix @ total_matrix

    # return total_matrix[:3, :3], total_matrix[:3, 3]
    return matrix


def _get_transformations(struct_oper):
    """
    Get transformation operation in terms of rotation matrix and
    translation for each operation ID in ``pdbx_struct_oper_list``.
    """
    transformation_dict = {}
    for index, id in enumerate(struct_oper["id"].as_array()):
        rotation_matrix = np.array(
            [
                [float(struct_oper[f"matrix[{i}][{j}]"][index]) for j in (1, 2, 3)]
                for i in (1, 2, 3)
            ]
        )
        translation_vector = np.array(
            [float(struct_oper[f"vector[{i}]"][index]) for i in (1, 2, 3)]
        )
        transformation_dict[id] = (rotation_matrix, translation_vector)
    return transformation_dict


def _parse_operation_expression(expression):
    """
    Get successive operation steps (IDs) for the given
    ``oper_expression``.
    Form the cartesian product, if necessary.
    """
    # Split groups by parentheses:
    # use the opening parenthesis as delimiter
    # and just remove the closing parenthesis
    expressions_per_step = expression.replace(")", "").split("(")
    expressions_per_step = [e for e in expressions_per_step if len(e) > 0]
    # Important: Operations are applied from right to left
    expressions_per_step.reverse()

    operations = []
    for expr in expressions_per_step:
        if "-" in expr:
            if "," in expr:
                for gexpr in expr.split(","):
                    if "-" in gexpr:
                        first, last = gexpr.split("-")
                        operations.append(
                            [str(id) for id in range(int(first), int(last) + 1)]
                        )
                    else:
                        operations.append([gexpr])
            else:
                # Range of operation IDs, they must be integers
                first, last = expr.split("-")
                operations.append([str(id) for id in range(int(first), int(last) + 1)])
        elif "," in expr:
            # List of operation IDs
            operations.append(expr.split(","))
        else:
            # Single operation ID
            operations.append([expr])

    # Cartesian product of operations
    return list(itertools.product(*operations))



import numpy as np
import molecularnodes as mn
from molecularnodes.utils import frame_mapper


def test_correct_1d():
    assert np.allclose(
        mn.utils.correct_periodic_1d(np.array((0.9, 0.1)), np.array((0.1, 0.9)), 1.0),
        np.array((1.1, -0.1)),
    )


def test_frame_mapper_basic():
    assert frame_mapper(10) == 10
    assert frame_mapper(0) == 0
    assert frame_mapper(-2) == 0


def test_frame_mapper_with_offset():
    assert frame_mapper(10, offset=2) == 8
    assert frame_mapper(2, offset=2) == 0
    assert frame_mapper(0, offset=5) == 0


def test_frame_mapper_with_subframes():
    assert frame_mapper(10, subframes=1) == 5
    assert frame_mapper(9, subframes=2) == 3


def test_frame_mapper_with_mapping():
    mapping = np.array([0, 0, 0, 1, 2])
    assert frame_mapper(1, mapping=mapping) == 0
    assert frame_mapper(3, mapping=mapping) == 1
    assert frame_mapper(3, mapping=list(mapping)) == 1


def test_frame_mapper_with_mapping_and_subframes():
    mapping = np.array([0, 0, 0, 1, 2])
    assert frame_mapper(5, subframes=1, mapping=mapping) == 0



from typing import Tuple
import bpy
from bpy.types import Material, ShaderNodeTree
from databpy.material import append_from_blend
from ..assets import MN_DATA_FILE
from .interface import TreeInterface, check_linked, remove_linked, socket

MATERIAL_NAMES = [
    "MN Default",
    "MN Flat Outline",
    "MN Squishy",
    "MN Transparent Outline",
    "MN Ambient Occlusion",
]


def append_material(name: str) -> bpy.types.Material:
    "Append a material from the MN_DATA_FILE."
    return append_from_blend(name, str(MN_DATA_FILE))


def add_all_materials() -> None:
    "Append all pre-defined materials from the MN_DATA_FILE."
    for name in MATERIAL_NAMES:
        append_material(name)


# class to interact with a bpy.types.Material node tree and change some of the default
# values of the nodes inside of it
class MaterialTreeInterface(TreeInterface):
    def __init__(self, material: Material):
        super().__init__()
        self._allowable_properties.add("material")
        if isinstance(material, str):
            material = bpy.data.materials[material]
        elif isinstance(material, Material):
            material = material
        elif isinstance(
            material, bpy.types.Material
        ):  # note this is equivalent class to the above.
            material = material
        else:
            raise ValueError(
                f"Material must be a string or a Material object. Found {type(material)}"
            )

        self.material: Material = material
        self.tree: ShaderNodeTree = self.material.node_tree  # type: ignore

    def _expose_all_inputs(self):
        for node in self.tree.nodes:
            if "Material Output" in node.name:
                continue
            for input in node.inputs:
                if input.is_unavailable:
                    continue
                input_name = input.name.lower().replace(" ", "_")
                node_name = (
                    (node.name if node.label == "" else node.label)
                    .lower()
                    .replace(" ", "_")
                )
                prop_name = f"{node_name}_{input_name}"

                if hasattr(input, "default_value"):
                    self._register_property(prop_name)
                    setattr(self.__class__, prop_name, socket(input))


def set_socket_material(
    socket: bpy.types.NodeSocketMaterial,
    mat: MaterialTreeInterface
    | bpy.types.Material
    | bpy.types.NodeSocketMaterial
    | str
    | None,
) -> None:
    remove_linked(socket)
    if mat is None:
        socket.default_value = None
    elif isinstance(mat, MaterialTreeInterface):
        socket.default_value = mat.material
    elif isinstance(mat, bpy.types.Material):
        socket.default_value = mat
    elif isinstance(mat, bpy.types.NodeSocketMaterial):
        socket.node.id_data.links.new(mat, socket)  # type: ignore
    elif isinstance(mat, str):
        mat = bpy.data.materials[mat]
        socket.default_value = mat
    else:
        raise TypeError("Invalid type for setting of a material: " + str(type(mat)))


def assign_material(
    node: bpy.types.GeometryNodeGroup,
    new_material: MaterialTreeInterface
    | bpy.types.Material
    | bpy.types.NodeSocketMaterial
    | str
    | None = "default",
) -> None:
    add_all_materials()

    if isinstance(new_material, str):
        if new_material not in bpy.data.materials:
            try:
                append_material(new_material)
            except Exception:
                try:
                    new_material = "MN " + new_material.title().strip()
                    append_material(new_material)
                except Exception:
                    raise ValueError(
                        f"Material {new_material} not found in this file of the included MN preset file."
                    )
    try:
        set_socket_material(
            socket=node.inputs["Material"],
            mat=new_material,
        )
    except KeyError:
        return "Material input not found on node."


def dynamic_material_interface(material: bpy.types.Material) -> MaterialTreeInterface:
    class_name = (
        f"DynamicMaterialInterface_{material.name.replace(' ', '_').replace('.', '_')}"
    )

    DynaicMaterialInterface = type(
        class_name,
        (MaterialTreeInterface,),
        {},
    )

    interface = DynaicMaterialInterface(material)
    interface._expose_all_inputs()
    return interface


def getset_material(socket: bpy.types.NodeSocketMaterial):
    def getter(self) -> MaterialTreeInterface | None:
        check_linked(socket)
        mat = getattr(socket, "default_value")
        if mat is None:
            return None
        else:
            interface = dynamic_material_interface(mat)
            return interface

    def setter(
        self,
        mat: MaterialTreeInterface
        | bpy.types.Material
        | bpy.types.NodeSocketMaterial
        | str
        | None,
    ) -> None:
        set_socket_material(socket, mat)

    return property(getter, setter)


class MaterialConstructor(MaterialTreeInterface):
    def __init__(self, material_name: str, **kwargs):
        super().__init__(append_material(material_name))
        self._expose_all_inputs()
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise ValueError(f"Material does not have property {key}")


class AmbientOcclusion(MaterialConstructor):
    def __init__(self, **kwargs):
        super().__init__("MN Ambient Occlusion", **kwargs)


class Default(MaterialConstructor):
    def __init__(self, **kwargs):
        super().__init__("MN Default", **kwargs)


class FlatOutline(MaterialConstructor):
    def __init__(self, **kwargs):
        super().__init__("MN Flat Outline", **kwargs)


class Squishy(MaterialConstructor):
    def __init__(self, **kwargs):
        super().__init__("MN Squishy", **kwargs)


class TransparentOutline(MaterialConstructor):
    def __init__(self, **kwargs):
        super().__init__("MN Transparent Outline", **kwargs)


class Material:
    """
    Note: if this is a nice route we could add fns that will create MN_materials here and remove them from the blend file.

    Materials:
    - AmbientOcclusion
    - Default  ( tried to replicate the defaults below )
    - FlatOutlineq
    - Squishy
    - TransparentOutline

    """

    @staticmethod
    def glass() -> bpy.types.Material:
        return create_material(
            name="MN Glass",  # Add a descriptive name
            base_color=(0.8, 0.9, 1.0, 0.2),
            transmission_weight=0.95,
            roughness=0.0,
            ior=1.45,
            specular_ior_level=0.6,  # Note: Renamed to "Specular IOR" in create_material
        )

    @staticmethod
    def gold() -> bpy.types.Material:
        return create_material(
            name="MN Gold",
            base_color=(1.0, 0.8, 0.2, 1.0),
            metallic=1.0,
            roughness=0.3,
            anisotropic=0.8,
            anisotropic_rotation=0.5,
        )

    @staticmethod
    def green_glow() -> bpy.types.Material:
        return create_material(
            name="MN GreenGlow",
            base_color=(0.0, 1.0, 0.0, 1.0),
            emission_strength=5.0,
            emission_color=(0.0, 1.0, 0.0, 1.0),
            metallic=0.0,
            roughness=0.2,
        )

    @staticmethod
    def holo() -> bpy.types.Material:
        return create_material(
            name="MN Holo",
            base_color=(0.2, 0.6, 1.0, 0.3),
            emission_strength=2.0,
            emission_color=(0.2, 0.6, 1.0, 1.0),
            transmission_weight=0.8,
            thin_film_thickness=1000,
        )

    @staticmethod
    def iridescent() -> bpy.types.Material:
        return create_material(
            name="MN Iridescent",
            base_color=(0.8, 0.8, 0.8, 0.3),
            metallic=0.8,
            transmission_weight=0.5,
            thin_film_thickness=1200,
            thin_film_ior=1.5,
            coat_weight=1.0,
        )

    @staticmethod
    def metallic() -> bpy.types.Material:
        return create_material(
            name="MN Metallic",
            base_color=(0.7, 0.7, 0.7, 1.0),
            metallic=1.0,
            roughness=0.1,
            specular_ior_level=0.9,
            coat_weight=1.0,
        )

    @staticmethod
    def neon() -> bpy.types.Material:
        return create_material(
            name="MN Neon",
            base_color=(0.0, 1.0, 0.8, 1.0),
            emission_strength=5.0,
            emission_color=(0.0, 1.0, 0.8, 1.0),
            metallic=0.8,
            roughness=0.1,
        )

    @staticmethod
    def new() -> bpy.types.Material:
        return create_material(name="MN Default 02")

    @staticmethod
    def pearl() -> bpy.types.Material:
        return create_material(
            name="MN Pearl",
            base_color=(0.9, 0.9, 0.9, 1.0),
            metallic=0.7,
            roughness=0.15,
            coat_weight=1.0,
            coat_ior=2.0,
            thin_film_thickness=500,
        )

    @staticmethod
    def subsurface() -> bpy.types.Material:
        return create_material(
            name="MN Subsurface",
            base_color=(1.0, 0.4, 0.4, 1.0),
            subsurface_weight=1.0,
            subsurface_radius=(1.0, 0.2, 0.1),
            subsurface_scale=0.5,
            emission_strength=0.3,
        )

    @staticmethod
    def toon() -> bpy.types.Material:
        return create_material(
            name="MN Toon",
            base_color=(0.2, 0.6, 1.0, 1.0),
            metallic=0,
            roughness=1.0,
            specular_ior_level=0.0,
            diffuse_roughness=1.0,
            coat_weight=0.2,
        )

    @staticmethod
    def velvet() -> bpy.types.Material:
        return create_material(
            name="MN Velvet",
            base_color=(0.5, 0.0, 0.2, 1.0),
            sheen_weight=1.0,
            sheen_roughness=0.3,
            sheen_tint=(1.0, 0.8, 0.9, 1.0),
            roughness=0.8,
        )

    @staticmethod
    def waxy() -> bpy.types.Material:
        return create_material(
            name="MN Waxy",
            base_color=(0.9, 0.87, 0.82, 1.0),
            subsurface_weight=0.3,
            subsurface_radius=(0.5, 0.4, 0.3),
            subsurface_scale=0.1,
            roughness=0.4,
            specular_ior_level=0.2,
            metallic=0.0,
            coat_weight=0.1,
            coat_roughness=0.3,
            sheen_weight=0.1,
            sheen_roughness=0.3,
        )


def create_material(
    name: str | None = None,
    base_color: Tuple[float, float, float, float] = (0.8, 0.8, 0.8, 0.05),
    metallic: float = 0.0,
    roughness: float = 0.2,
    ior: float = 1.45,
    alpha: float = 1.0,
    normal: Tuple[float, float, float] = (0.0, 0.0, 0.0),
    weight: float = 0.0,
    diffuse_roughness: float = 0.0,
    subsurface_weight: float = 0.0,
    subsurface_radius: Tuple[float, float, float] = (1.0, 0.2, 0.1),
    subsurface_scale: float = 0.05,
    subsurface_ior: float = 1.4,
    subsurface_anisotropy: float = 0.0,
    specular_ior_level: float = 0.5,
    specular_tint: Tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0),
    anisotropic: float = 0.0,
    anisotropic_rotation: float = 0.0,
    tangent: Tuple[float, float, float] = (0.0, 0.0, 0.0),
    transmission_weight: float = 0.0,
    coat_weight: float = 0.0,
    coat_roughness: float = 0.03,
    coat_ior: float = 1.5,
    coat_tint: Tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0),
    coat_normal: Tuple[float, float, float] = (0.0, 0.0, 0.0),
    sheen_weight: float = 0.0,
    sheen_roughness: float = 0.5,
    sheen_tint: Tuple[float, float, float, float] = (0.5, 0.5, 0.5, 1.0),
    emission_color: Tuple[float, float, float, float] = (0.0, 0.0, 0.0, 1.0),
    emission_strength: float = 0.0,
    thin_film_thickness: float = 0.0,
    thin_film_ior: float = 1.3,
) -> bpy.types.Material:
    """Create a Blender material from provided shader properties."""

    key_map = {
        "Base Color": base_color,
        "Metallic": metallic,
        "Roughness": roughness,
        "IOR": ior,
        "Alpha": alpha,
        "Normal": normal,
        "Weight": weight,
        "Diffuse Roughness": diffuse_roughness,
        "Subsurface Weight": subsurface_weight,
        "Subsurface Radius": subsurface_radius,
        "Subsurface Scale": subsurface_scale,
        "Subsurface IOR": subsurface_ior,
        "Subsurface Anisotropy": subsurface_anisotropy,
        "Specular IOR Level": specular_ior_level,
        "Specular Tint": specular_tint,
        "Anisotropic": anisotropic,
        "Anisotropic Rotation": anisotropic_rotation,
        "Tangent": tangent,
        "Transmission Weight": transmission_weight,
        "Coat Weight": coat_weight,
        "Coat Roughness": coat_roughness,
        "Coat IOR": coat_ior,
        "Coat Tint": coat_tint,
        "Coat Normal": coat_normal,
        "Sheen Weight": sheen_weight,
        "Sheen Roughness": sheen_roughness,
        "Sheen Tint": sheen_tint,
        "Emission Color": emission_color,
        "Emission Strength": emission_strength,
        "Thin Film Thickness": thin_film_thickness,
        "Thin Film IOR": thin_film_ior,
    }

    if name is None:
        name = f"material_func_{id(key_map)}"

    if name in bpy.data.materials:
        return bpy.data.materials[name]

    mat = bpy.data.materials.new(name)

    if bpy.app.version < (5, 0, 0):
        mat.use_nodes = True

    tree = mat.node_tree
    if tree is None:
        raise RuntimeError("Material does not have a node tree")
    try:
        bsdf = mat.node_tree.nodes["Principled BSDF"]  # type: ignore
    except KeyError:
        output = tree.nodes.new("ShaderNodeOutputMaterial")
        bsdf = tree.nodes.new("ShaderNodeBsdfPrincipled")
        bsdf.location = (-200, 0)
        tree.links.new(bsdf.outputs["BSDF"], output.inputs["Surface"])

    for input_socket in bsdf.inputs:
        if input_socket.name in key_map:
            input_socket.default_value = key_map[input_socket.name]

    return mat



from typing import List, Sequence
import bpy
from bpy.types import Node  # type: ignore
from mathutils import Vector  # type: ignore
from . import nodes
from .arrange import arrange_tree
from .interface import (
    TreeInterface,
    input_named_attribute,
    socket,
)
from .material import (
    assign_material,
    getset_material,
)
from .nodes import (
    NODE_SPACING,
    final_join,
    insert_before,
    loc_between,
)
from .styles import StyleBase


def insert_set_color(
    node: bpy.types.Node,
    color: str | Sequence[float] = [0.3, 0.3, 0.3, 1.0],
) -> bpy.types.GeometryNodeGroup:
    """
    Add a set color node to the tree and connect it to the given socket
    """
    _tree = node.id_data
    node_sc: bpy.types.GeometryNodeGroup = insert_before(node, "Set Color")  # type: ignore

    if isinstance(color, str):
        if color.lower() in ["default", "common"]:
            node_cc = insert_before(node_sc.inputs["Color"], "Color Common")
            node_car: bpy.types.GeometryNodeGroup = insert_before(  # type: ignore
                node_cc.inputs["Carbon"], "Color Attribute Random"
            )
            return node_car
        elif color.lower() == "plddt":
            insert_before(node_sc.inputs["Color"], "Color pLDDT")
        else:
            input_named_attribute(node_sc.inputs["Color"], color, "FLOAT_COLOR")
    else:
        node_sc.inputs["Color"].default_value = color  # type: ignore
    return node_sc


def insert_animate_frames(
    node: bpy.types.GeometryNode, frames: bpy.types.Collection | str
) -> bpy.types.Node:
    """
    Add an animate frames node to the tree and connect it to the given socket
    """
    node.location += Vector([NODE_SPACING, 0])
    tree = node.id_data
    node_af: bpy.types.GeometryNodeGroup = insert_before(node, "Animate Frames")  # type: ignore
    if isinstance(frames, bpy.types.Collection):
        node_af.inputs["Frames"].default_value = frames  # type: ignore
    elif isinstance(frames, str):
        node_af.inputs["Frames"].default_value = bpy.data.collections[frames]  # type: ignore
    else:
        raise ValueError(
            f"Frames must be a string or a Collection, not {type(frames)=}"
        )

    node_an = nodes.add_custom(
        tree,
        "Animate Value",
        location=node_af.location + Vector([-NODE_SPACING, -NODE_SPACING / 1.5]),
    )
    node_an.inputs["Value Max"].default_value = float(len(frames.objects) - 1)  # type: ignore
    tree.links.new(node_an.outputs[0], node_af.inputs["Frame"])

    return node_af


def add_style_branch(
    tree: bpy.types.GeometryNodeTree,
    style: str | bpy.types.GeometryNodeTree | StyleBase,
    color: str | None = None,
    selection: str | None = None,
    material: bpy.types.Material | str | None = None,
    frames: bpy.types.Collection | str | None = None,
    name: str | None = None,
) -> bpy.types.GeometryNodeGroup:
    """
    Add a style branch to the tree.
    """
    _frame = True
    link = tree.links.new
    input = nodes.get_input(tree)
    node_join = final_join(tree)

    current_min_y = min(node.location[1] for node in tree.nodes)
    ypos = current_min_y - 200
    xpos = loc_between(input, node_join, 0.75)[0]

    if isinstance(style, str):
        style_name = nodes.styles_mapping[style]
        nodes.append(style_name)
    elif isinstance(style, StyleBase):
        style_name = nodes.styles_mapping[style.style]
        nodes.append(style_name)
    elif isinstance(style, bpy.types.GeometryNodeTree):
        style_name = style.name
    else:
        raise ValueError(
            f"Style must be a string or a GeometryNodeTree, not {type(style)=}"
        )

    node_style = nodes.add_custom(
        group=tree,
        name=style_name,
        location=[xpos, ypos],
    )

    link(
        input.outputs[0],
        node_style.inputs[0],
    )
    link(
        node_style.outputs[0],
        node_join.inputs[0],
    )

    for nodelink in node_join.inputs[0].links:  # type: ignore
        if nodelink.from_socket.node == nodes.get_input(tree):  # type: ignore
            tree.links.remove(nodelink)

    # Apply style modifications
    if material:
        assign_material(node_style, material)
    if selection:
        input_named_attribute(node_style.inputs["Selection"], selection, "BOOLEAN")
    if color:
        insert_set_color(node_style, color)
    if frames:
        insert_animate_frames(node_style, frames)

    arrange_tree(tree)

    if isinstance(style, StyleBase):
        style.update_style_node(node_style)

    if name is not None:
        node_style.label = name
    else:
        node_style.label = node_style.name

    return node_style


def get_final_style_nodes(
    tree: bpy.types.GeometryNodeTree,
) -> List[bpy.types.Node | None]:
    """
    Get the final style nodes in the tree.
    """
    links: bpy.types.NodeLinks = final_join(tree).inputs[0].links  # type: ignore
    return [
        link.from_socket.node  # type: ignore
        for link in reversed(links)
        if link.from_socket.node.name.startswith("Style")  # type: ignore
    ]


class GeometryNodeInterFace(TreeInterface):
    """
    Interface for the geometry nodes in the tree.
    """

    def __init__(self, node: bpy.types.Node) -> None:
        super().__init__()
        self.tree: bpy.types.NodeTree = node.id_data
        self._nodes = []

    def remove(self) -> None:
        """
        Cleanup when this instance is explicitly deleted.
        """
        for node in self._nodes:
            self.tree.nodes.remove(node)
        arrange_tree(self.tree)

    def _expose_options(self, node: bpy.types.Node) -> None:
        self._nodes.append(node)
        for input in node.inputs:
            if not hasattr(input, "default_value"):
                continue

            # Create property name from node and input names
            prop_name = (
                "_".join([node.name.split(".")[0], input.name])
                .lower()
                .replace(" ", "_")
                .removeprefix("style_")
                .removeprefix("ribbon_")
                .removeprefix("surface_")
                .removeprefix("ball_and_stick_")
                .removeprefix("cartoon_")
                .removeprefix("spheres_")
                .removeprefix("density_surface_")
                .removeprefix("density_iso_surface_")
                .removeprefix("density_wire_")
            )
            if isinstance(input, bpy.types.NodeSocketMaterial):
                prop = getset_material(input)
            else:
                prop = socket(input)
            self._register_property(prop_name)
            setattr(self.__class__, prop_name, prop)


def create_style_interface(node: Node, linked: bool = True) -> GeometryNodeInterFace:
    """
    Dynamically create a StyleInterface class with exposed options.
    """
    class_name = f"DynamicStyleInterface_{node.name.replace('Style ', '')}"

    # Create the dynamic class with a unique name
    DynamicInterface = type(class_name, (GeometryNodeInterFace,), {})

    # Pre-expose the options for the main node
    interface = DynamicInterface(node)
    interface._expose_options(node)

    if not linked:
        return interface

    # Pre-expose options for linked nodes
    for input in node.inputs:
        if input.is_linked:
            linked_node = input.links[0].from_socket.node  # type: ignore
            interface._expose_options(linked_node)

    return interface


class StyleManager(List):
    def clear(self) -> None:
        while len(self) > 0:
            self.pop().remove()


def style_interfaces_from_tree(
    tree: bpy.types.GeometryNodeTree,
) -> StyleManager[GeometryNodeInterFace]:
    return StyleManager(
        create_style_interface(node) for node in get_final_style_nodes(tree)
    )



import os
from abc import ABCMeta
from pathlib import Path
from typing import Union
import databpy
import numpy as np
from ... import blender as bl
from ..base import EntityType, MolecularEntity
from .annotations import DensityAnnotationManager


class Density(MolecularEntity, metaclass=ABCMeta):
    """
    Abstract base class for molecular density objects.

    """

    def __init__(self, file_path: Union[str, Path]):
        super().__init__()
        self._entity_type = EntityType.DENSITY
        self.file_path: Path = bl.path_resolve(file_path)
        self.file_vdb: str
        self.annotations = DensityAnnotationManager(self)

    def named_attribute(self, name: str, evaluate: bool = True) -> np.ndarray:
        obj = bl.mesh.evaluate_using_mesh(self.object)
        return databpy.named_attribute(obj, name, evaluate=True)

    def path_to_vdb(self, file: str, center: bool = False, invert: bool = False):
        """
        Convert a file path to a corresponding VDB file path.

        Parameters
        ----------
        file : str
            The path of the original file.
        center : bool, default=False
            If True, the density will be centered at the origin.
        invert : bool, default=False
            If True, the density values will be inverted.

        Returns
        -------
        str
            The path of the corresponding VDB file.
        """
        # Set up file paths
        folder_path = os.path.dirname(file)
        name = os.path.basename(file).split(".")[0]
        name += "_center" if center else ""
        name += "_invert" if invert else ""
        file_name = name + ".vdb"
        file_path = os.path.join(folder_path, file_name)
        return file_path
